import importlib

from .eig import EIG
from .pflow import PFLOW
from .tds import TDS


def get_command(all_pkg, hook):
    """
    Collect the command-line interface names by querying ``hook`` in ``all_pkg``

    Parameters
    ----------
    all_pkg: list
        list of package files
    hook: str
        A variable where the command is stored. ``__cli__`` by default.
    Returns
    -------
    list
    """
    ret = []
    for r in all_pkg:
        module = importlib.import_module(__name__ + '.' + r.lower())
        ret.append(getattr(module, hook))
    return ret


__all__ = ['PFLOW', 'TDS', 'EIG']
__cli__ = get_command(__all__, '__cli__')



import logging
import os

logger = logging.getLogger(__name__)


class FileMan(object):
    """Define a File Manager class for PowerSystem"""

    def __init__(self,
                 case,
                 input_format=None,
                 input_path=None,
                 addfile=None,
                 config=None,
                 no_output=False,
                 dynfile=None,
                 dump_raw=None,
                 output_format=None,
                 output_path='',
                 output=None,  # base file name for the output
                 pert=None,
                 **kwargs):
        """
        Initialize the output file names.
        For inputs, all absolute paths will be respected; and all relative paths are relative to `input_path`.

        case: must be full path to case

        dump_raw: desired simulation result file name

        output: desired name for format conversion output

        input_path: default path for input files that only contains file name. If `input_path` is not provided,
                    it will be derived from the path of `case`.

        output_path: path for output files. Default to current working directory where `andes` is invoked.
        """
        if isinstance(input_format, str):
            self.input_format = input_format.lower()
        else:
            self.input_format = None
        if isinstance(output_format, str):
            self.output_format = output_format.lower()
        else:
            self.output_format = None

        if input_path is not None:
            logger.debug('input_path provided as {}.'.format(input_path))
            self.case_path = input_path
        else:
            self.case_path = os.getcwd()  # default to current directory

        if os.path.isabs(case):
            self.case = case
        else:
            self.case = self.get_fullpath(case)
            logger.debug(self.case)

        # update `self.case_path` if `case` contains a path
        self.case_path, self.fullname = os.path.split(self.case)

        # `self.name` is the name part without extension
        self.name, self.ext = os.path.splitext(self.fullname)

        self.addfile = self.get_fullpath(addfile)
        self.pert = self.get_fullpath(pert)
        self.dynfile = self.get_fullpath(dynfile)
        self.config = self.get_fullpath(config)
        self.add_format = None

        # use the path where andes is executed as the default output path
        self.output_path = os.getcwd() if not output_path else output_path
        if no_output:
            self.no_output = True
            self.output = None
            self.lst = None
            self.eig = None
            self.dat = None
            self.dump_raw = None
            self.prof = None
        else:
            self.no_output = False
            if not output:
                output = add_suffix(self.name, 'out')
            if not dump_raw:
                dump_raw = add_suffix(self.name, 'dm')
            prof = add_suffix(self.name, 'prof')
            eig = add_suffix(self.name, 'eig')

            self.lst = os.path.join(self.output_path, output + '.lst')
            self.dat = os.path.join(self.output_path, output + '.dat')
            self.output = os.path.join(self.output_path, output + '.txt')

            self.eig = os.path.join(self.output_path, eig + '.txt')
            self.dump_raw = os.path.join(self.output_path, dump_raw + '.dm')
            self.prof = os.path.join(self.output_path, prof + '.txt')

    def get_fullpath(self, fullname=None, relative_to=None):
        """
        Return the original full path if full path is specified, otherwise
        search in the case file path
        """
        # if is an empty path
        if not fullname:
            return fullname

        isabs = os.path.isabs(fullname)

        path, name = os.path.split(fullname)

        if not name:  # path to a folder
            return None
        else:  # path to a file
            if isabs:
                return fullname
            else:
                return os.path.join(self.case_path, path, name)


def add_suffix(fullname, suffix):
    """ Add suffix to a full file name"""
    name, ext = os.path.splitext(fullname)
    return name + '_' + suffix + ext



#!/usr/bin/env python3

# ANDES, a power system simulation tool for research.
#
# Copyright 2015-2017 Hantao Cui
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Andes main entry points
"""

import cProfile
import glob
import io
import logging
import os
import pathlib
import platform
import pprint
import pstats
import sys
from argparse import ArgumentParser
from multiprocessing import Process
from subprocess import call
from time import sleep, strftime

from andes import filters
from andes import routines
from andes import utils
from andes.system import PowerSystem
from andes.utils import elapsed, misc

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def config_logger(logger=None, name='andes', log_file='andes.log', log_path='', stream=True, file=False,
                  stream_level=logging.INFO, file_level=logging.DEBUG):
    """
    Configure a logger for the andes package with options for a `FileHandler`
    and a `StreamHandler`. This function is called at the beginning of
    ``andes.main.main()``.

    Parameters
    ----------
    name : str, optional
        Base logger name, ``andes`` by default. Changing this
        parameter will affect the loggers in modules and
        cause unexpected behaviours.
    log_file : str, optional
        Logg file name for `FileHandler`, ``'andes.log'`` by default.
        If ``None``, the `FileHandler` will not be created.
    log_path : str, optional
        Path to store the log file. By default, the path is generated by
        get_log_dir() in utils.misc.
    stream : bool, optional
        Create a `StreamHandler` for `stdout` if ``True``.
        If ``False``, the handler will not be created.
    stream_level : {10, 20, 30, 40, 50}, optional
        `StreamHandler` verbosity level.

    Returns
    -------
    None

    """
    if not logger:
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

    if not len(logger.handlers):
        if stream is True:
            sh_formatter = logging.Formatter('%(message)s')
            sh = logging.StreamHandler()

            sh.setFormatter(sh_formatter)
            sh.setLevel(stream_level)
            logger.addHandler(sh)

        # file handler for level DEBUG and up
        if file is True and (log_file is not None):
            log_full_path = os.path.join(log_path, log_file)
            fh_formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            fh = logging.FileHandler(log_full_path)
            fh.setLevel(file_level)
            fh.setFormatter(fh_formatter)
            logger.addHandler(fh)
            logger.debug('Logging to file {}'.format(log_full_path))

        globals()['logger'] = logger


def preamble():
    """
    Log the Andes command-line preamble at the `logging.INFO` level

    Returns
    -------
    None
    """
    from andes import __version__ as version
    logger.info('ANDES {ver} (Build {b}, Python {p} on {os})'
                .format(ver=version[:5], b=version[-8:],
                        p=platform.python_version(),
                        os=platform.system()))
    try:
        username = os.getlogin() + ', '
    except OSError:
        username = ''

    logger.info('Session: {}{}'.format(username,
                                       strftime("%m/%d/%Y %I:%M:%S %p")))
    logger.info('')


def cli_parser():
    """
    Construct a CLI argument parser and return the parsed arguments.

    Returns
    -------
    ArgumentParser
        An argument parser for parsing command-line arguments
    """
    parser = ArgumentParser()
    parser.add_argument('filename', help='Case file name', nargs='*')

    # general options
    general_group = parser.add_argument_group('General options')
    general_group.add_argument('-r', '--routine', choices=routines.__cli__,
                               help='Routine to run', nargs='*',
                               default=['pflow'], )
    general_group.add_argument('--edit-config', help='Quick edit of the config file',
                               default='', nargs='?', type=str)
    general_group.add_argument('--license', action='store_true', help='Display software license')

    # I/O
    io_group = parser.add_argument_group('I/O options', 'Optional arguments for managing I/Os')
    io_group.add_argument('-p', '--path', help='Path to case files', type=str, default='', dest='input_path')
    io_group.add_argument('-a', '--addfile', help='Additional files used by some formats.')
    io_group.add_argument('-D', '--dynfile', help='Additional dynamic file in dm format.')
    io_group.add_argument('-P', '--pert', help='Perturbation file path', default='')
    io_group.add_argument('-d', '--dump-raw', help='Dump RAW format case file.')
    io_group.add_argument('-n', '--no-output', help='Force no output of any '
                                                    'kind',
                          action='store_true')
    io_group.add_argument('-o', '--output_path', help='Output path prefix', type=str, default='')
    io_group.add_argument('-C', '--clean', help='Clean output files', action='store_true')

    config_exclusive = parser.add_mutually_exclusive_group()
    config_exclusive.add_argument('--load-config', help='path to the rc config to load',
                                  dest='config')
    config_exclusive.add_argument('--save-config', help='save configuration to file name',
                                  nargs='?', type=str, default='')

    # helps and documentations
    group_help = parser.add_argument_group('Help and documentation',
                                           'Optional arguments for usage, model and config documentation')
    group_help.add_argument(
        '-g', '--group', help='Show the models in the group.')
    group_help.add_argument(
        '-q', '--quick-help', help='Show a quick help of model format.')
    group_help.add_argument(
        '-c',
        '--category',
        help='Show model names in the given category.')
    group_help.add_argument(
        '-l',
        '--model-list',
        help='Show a full list of all models.',
        action='store_true')
    group_help.add_argument(
        '-f',
        '--model-format',
        help='Show the format definition of models.', type=str)
    group_help.add_argument(
        '-Q',
        '--model-var',
        help='Show the definition of variables <MODEL.VAR>.')
    group_help.add_argument(
        '--config-option', help='Show a quick help of a config option <CONFIG.OPTION>')
    group_help.add_argument(
        '--help-config',
        help='Show help of the <CONFIG> class. Use ALL for all configs.')
    group_help.add_argument(
        '-s',
        '--search',
        help='Search for models that match the pattern.')
    group_help.add_argument('-e', '--data_example', help='print example parameter of a given model')

    # simulation control
    sim_options = parser.add_argument_group('Simulation control options',
                                            'Overwrites the simulation configs')
    sim_options.add_argument(
        '--dime', help='Specify DiME streaming server address and port')
    sim_options.add_argument(
        '--tf', help='End time of time-domain simulation', type=float)

    # developer options
    dev_group = parser.add_argument_group('Developer options', 'Options for developer debugging')
    dev_group.add_argument(
        '-v',
        '--verbose',
        help='Program logging level.'
             'Available levels are 10-DEBUG, 20-INFO, 30-WARNING, '
             '40-ERROR or 50-CRITICAL. The default level is 20-INFO',
        type=int, default=20, choices=(10, 20, 30, 40, 50))
    dev_group.add_argument(
        '--profile', action='store_true', help='Enable Python cProfiler')
    dev_group.add_argument(
        '--ncpu', help='Number of parallel processes', type=int, default=0)
    dev_group.add_argument(
        '--show-data', type=str, help='Show model data converted to system base', nargs='*'
    )
    dev_group.add_argument(
        '-x', '--exit', help='Exit before running routine', action='store_true', dest='exit_now'
    )

    return parser


def cli_new(parser):
    """
    Do parse with the provided CLI parser

    Parameters
    ----------
    parser : ArgumentParser
        An ArgumentParser instance for parsing the command line arguments

    Returns
    -------
    Namespace
        A namespace containing the parsed arguments

    """
    args = parser.parse_args()
    return args


def andeshelp(group=None, category=None, model_list=None, model_format=None, model_var=None,
              quick_help=None, help_option=None, help_config=None, export='plain', data_example=None, **kwargs):
    """
    Print the requested help and documentation to stdout.

    Parameters
    ----------
    group : None or str
        Name of a group whose model names will be printed

    category : None or str
        Name of a category whose models will be printed

    model_list : bool
        If ``True``, print the full model list.

    model_format : None or str
        Names of models whose parameter definitions will be printed. Model
        names are separated by comma without space.

    model_var : None or str
        A pair of model name and parameter name separated by dot. For
        example, ``Bus.voltage`` stands for the ``voltage`` parameter of
        ``Bus``.

    quick_help : None or str
        Name of a model to print a quick help of parameter definitions.

    help_option : None or str
        A pair of config name and option name separated by dot. For example,
        ``System.sparselib`` stands for the ``sparselib`` option of
        the ``System`` config.

    help_config : None or str
        Config names whose option definitions will be printed. Configs are
        separated by comma without space. For example, ``System,Pflow``. In
        the naming convention, the first letter is captialized.

    export : None or {'plain', 'latex'}
        Formatting style available in plain text or LaTex format. This option
        has not been implemented.

    data_example : str, optional
        Print example data for a specified model. The data can be used to construct
        a dm data file.

    kwargs : None or dict
        Other keyword arguments

    Returns
    -------
    bool
        True if any help function is executed.

    Notes
    -----
    The outputs can be written to a text file using shell command,
    for example, ::

        andes -q Bus > bus_help.txt

    """
    out = []

    if not (group or category or model_list or model_format
            or model_var or quick_help or help_option or help_config or data_example):
        return False

    from andes.models import all_models_list

    if category:
        raise NotImplementedError

    if model_list:
        raise NotImplementedError

    system = PowerSystem()

    # print example data
    if data_example is not None:
        out.append(system.get_data_example(data_example))

    if model_format:
        if model_format.lower() == 'all':
            model_format = all_models_list
        else:
            model_format = model_format.split(',')

        for item in model_format:
            if item not in all_models_list:
                logger.warning('Model <{}> does not exist.'.format(item))
                model_format.remove(item)

        if len(model_format) > 0:
            for item in model_format:
                out.append(system.__dict__[item].doc(export=export))

    if model_var:
        model_var = model_var.split('.')

        if len(model_var) == 1:
            logger.error('Model and parameter not separated by dot.')

        elif len(model_var) > 2:
            logger.error('Model parameter not specified correctly.')

        else:
            dev, var = model_var
            if not hasattr(system, dev):
                logger.error('Model <{}> does not exist.'.format(dev))
            else:
                if var not in system.__dict__[dev]._data.keys():
                    logger.error(
                        'Model <{}> does not have parameter <{}>.'.format(
                            dev, var))
                else:
                    c1 = system.__dict__[dev]._descr.get(var, 'no Description')
                    c2 = system.__dict__[dev]._data.get(var)
                    c3 = system.__dict__[dev]._units.get(var, 'no Unit')
                    out.append('{}: {}, default = {:g} {}'.format(
                        '.'.join(model_var), c1, c2, c3))

    if group:
        group_dict = {}
        match = []

        for model in all_models_list:
            g = system.__dict__[model]._group
            if g not in group_dict:
                group_dict[g] = []
            group_dict[g].append(model)

        if group.lower() == 'all':
            match = sorted(list(group_dict.keys()))

        else:
            group = [group]
            # search for ``group`` in all group names and store in ``match``
            for item in group_dict.keys():
                if group[0].lower() in item.lower():
                    match.append(item)

        # if group name not found
        if len(match) == 0:
            out.append('Group <{:s}> not found.'.format(group[0]))

        for item in match:
            group_models = sorted(list(group_dict[item]))
            out.append('<{:s}>'.format(item))
            out.append(', '.join(group_models))
            out.append('')

    if quick_help:
        if quick_help not in all_models_list:
            out.append('Model <{}> does not exist.'.format(quick_help))
        else:
            out.append(system.__dict__[quick_help].doc(export=export))

    if help_option:
        raise NotImplementedError

    if help_config:

        all_config = ['system', 'pflow', 'tds', 'eig']

        if help_config.lower() == 'all':
            help_config = all_config

        else:
            help_config = help_config.split(',')

            for item in help_config:
                if item.lower() not in all_config:
                    logger.warning('Config <{}> does not exist.'.format(item))
                    help_config.remove(item)

        if len(help_config) > 0:
            for item in help_config:
                if item == 'system':
                    out.append(system.config.doc(export=export))
                else:
                    out.append(system.__dict__[item.lower()].config.doc(
                        export=export))

    print('\n'.join(out))  # NOQA

    return True


def edit_conf(edit_config='', load_config=None, **kwargs):
    """
    Edit the Andes config file which occurs first in the search path.

    Parameters
    ----------
    edit_config : bool
        If ``True``, try to open up an editor and edit the config file.
        Otherwise returns.

    load_config : None or str, optional
        Path to the config file, which will be placed to the first in the
        search order.

    kwargs : dict
        Other keyword arguments.

    Returns
    -------
    bool
        ``True`` is a config file is found and an editor is opened. ``False``
        if ``edit_config`` is False.

    """
    ret = False

    # no `edit-config` supplied
    if edit_config == '':
        return ret

    conf_path = misc.get_config_load_path(load_config)

    if conf_path is not None:
        logger.info('Editing config file {}'.format(conf_path))

        if edit_config is None:
            # use the following default editors
            if platform.system() == 'Linux':
                editor = os.environ.get('EDITOR', 'gedit')
            elif platform.system() == 'Darwin':
                editor = os.environ.get('EDITOR', 'vim')
            elif platform.system() == 'Windows':
                editor = 'notepad.exe'
        else:
            # use `edit_config` as default editor
            editor = edit_config

        call([editor, conf_path])
        ret = True

    else:
        logger.info('Config file does not exist. Save config with \'andes '
                    '--save-config\'')
        ret = True

    return ret


def remove_output(clean=False, **kwargs):
    """
    Remove the outputs generated by Andes, including power flow reports
    ``_out.txt``, time-domain list ``_out.lst`` and data ``_out.dat``,
    eigenvalue analysis report ``_eig.txt``.

    Parameters
    ----------
    clean : bool
        If ``True``, execute the function body. Returns otherwise.

    kwargs : dict
        Other keyword arguments

    Returns
    -------
    bool
        ``True`` is the function body executes with success. ``False``
        otherwise.
    """
    if not clean:
        return False

    found = False
    cwd = os.getcwd()

    for file in os.listdir(cwd):
        if file.endswith('_eig.txt') or \
                file.endswith('_out.txt') or \
                file.endswith('_out.lst') or \
                file.endswith('_out.dat') or \
                file.endswith('_prof.txt'):
            found = True
            try:
                os.remove(file)
                logger.info('<{:s}> removed.'.format(file))
            except IOError:
                logger.error('Error removing file <{:s}>.'.format(file))
    if not found:
        logger.info('no output found in the working directory.')

    return True


def search(**kwargs):
    """
    Search for models whose names matches the given pattern. Print the
    results to stdout.

    .. deprecated :: 1.0.0
        `search` will be moved to ``andeshelp`` in future versions.

    Parameters
    ----------
    search : str
        Partial or full name of the model to search for

    kwargs : dict
        Other keyword arguments.

    Returns
    -------
    list
        The list of model names that match the given pattern.
    """

    from andes.models import all_models
    out = []

    search_str = kwargs.get('search', None)
    if not search_str:
        return out

    keys = sorted(list(all_models.keys()))

    for key in keys:
        vals = all_models[key]
        val = list(vals.keys())
        val = sorted(val)

        for item in val:
            if search_str.lower() in item.lower():
                out.append(key + '.' + item)

    if out:
        print('Search result: <file.model> containing <{}>'.format(search_str))
        print(' '.join(out))
    else:
        print('No model containing <{:s}> found'.format(search_str))

    return out


def save_config(save_config='', **kwargs):
    """
    Save the Andes config to a file at the path specified by ``save_config``.
    The save action will not run if `save_config = ''`.

    Parameters
    ----------
    save_config : None or str, optional, ('' by default)
        Path to the file to save the config file. If the path is an emtpy
        string, the save action will not run. Save to
        `~/.andes/andes.conf` if ``None``.

    kwargs : dict, optional
        Other keyword arguments

    Returns
    -------
    bool
        ``True`` is the save action is run. ``False`` otherwise.
    """
    ret = False
    cf_path = save_config

    # no ``--save-config ``
    if cf_path == '':
        return ret

    if cf_path is None:
        cf_path = 'andes.conf'
        home = str(pathlib.Path.home())

        path = os.path.join(home, '.andes')
        if not os.path.exists(path):
            os.makedirs(path)

        cf_path = os.path.join(path, cf_path)

    ps = PowerSystem()
    ps.dump_config(cf_path)
    ret = True

    return ret


def main(args=None):
    """
    The main function of the Andes command-line tool.

    This function executes the following workflow:

     * Parse the command line inputs
     * Show the tool preamble
     * Output the requested helps, edit/save configs or remove outputs. Exit
       the main program if any of the above is executed
     * Process the input files and call ``main.run()`` using single- or
       multi-processing
     * Show the execution time and exit

    Returns
    -------
    None
    """
    t0, _ = elapsed()

    # parser command line arguments
    if args is None:
        parser = cli_parser()
        args = vars(cli_new(parser))
    elif not isinstance(args, dict):
        args = vars(args)

    # configure stream handler verbose level
    config_logger(log_path=misc.get_log_dir(), file=True, stream=True,
                  stream_level=args.get('verbose', logging.INFO))

    # show preamble
    preamble()

    # ----- Debug only -----
    # logger.debug('command line arguments:')
    # logger.debug(pprint.pformat(args))
    # ----------------------

    if andeshelp(**args) or search(**args) or edit_conf(**args) or remove_output(**args) or save_config(**args):
        return

    # process input files
    filename = args.get('filename', ())
    if isinstance(filename, str):
        filename = [filename]

    if len(filename) == 0:
        logger.info('error: no input file. Try \'andes -h\' for help.')

    # preprocess cli args
    path = args.get('input_path', os.getcwd())
    ncpu = args.get('ncpu', 0)
    if ncpu == 0 or ncpu > os.cpu_count():
        ncpu = os.cpu_count()

    cases = []

    for file in filename:
        # use absolute path for cases which will be respected by FileMan
        full_paths = os.path.abspath(os.path.join(path, file))
        found = glob.glob(full_paths)
        if len(found) == 0:
            logger.info('error: file {} does not exist.'.format(full_paths))
        else:
            cases += found

    # remove folders and make cases unique
    cases = list(set(cases))
    valid_cases = []
    for case in cases:
        if os.path.isfile(case):
            valid_cases.append(case)

    logger.debug('Found files: ' + pprint.pformat(valid_cases))

    if len(valid_cases) <= 0:
        pass
    elif len(valid_cases) == 1:
        run(valid_cases[0], **args)
    else:
        # set verbose level for multi processing
        logger.info('Processing {} jobs on {} CPUs'.format(len(valid_cases), ncpu))
        logger.handlers[1].setLevel(logging.WARNING)

        # start processes
        jobs = []
        for idx, file in enumerate(valid_cases):
            args['pid'] = idx
            job = Process(
                name='Process {0:d}'.format(idx),
                target=run,
                args=(file,),
                kwargs=args)
            jobs.append(job)
            job.start()

            start_msg = 'Process {:d} <{:s}> started.'.format(idx, file)
            print(start_msg)
            logger.debug(start_msg)

            if (idx % ncpu == ncpu - 1) or (idx == len(valid_cases) - 1):
                sleep(0.1)
                for job in jobs:
                    job.join()
                jobs = []

        # restore command line output when all jobs are done
        logger.handlers[1].setLevel(logging.INFO)

    t0, s0 = elapsed(t0)

    if len(valid_cases) == 1:
        logger.info('-> Single process finished in {:s}.'.format(s0))
    elif len(valid_cases) >= 2:
        logger.info('-> Multiple processes finished in {:s}.'.format(s0))

    return


def run(case, routine=None, profile=False, dump_raw=False, pid=-1, show_data=None, exit_now=False, **kwargs):
    """
    Entry function to run a single case study. This function executes the
    following workflow:

     * Turn on cProfile if requested
     * Populate a ``PowerSystem`` object
     * Parse the input files using filters
     * Dump the case file is requested
     * Set up the system
     * Run the specified routine(s)

    Parameters
    ----------
    case : str
        Path to the case file

    profile : bool, optional
        Enable cProfile if ``True``.

    dump_raw : ``False`` or str, optional
        Path to the file to dump the system parameters in the dm format

    routine : Iterable, optional
        A list of routines to run in sequence (``('pflow')`` as default)

    pid : idx, optional
        Process idx of the current run

    kwargs : dict, optional
        Other keyword arguments

    Returns
    -------
    PowerSystem
        Andes PowerSystem object

    """
    t0, _ = elapsed()

    # enable profiler if requested
    pr = cProfile.Profile()
    if profile is True:
        pr.enable()

    system = PowerSystem(case, **kwargs)

    # guess format and parse data
    if not filters.guess(system):
        return

    if not filters.parse(system):
        return

    # dump system as raw file if requested
    if dump_raw:
        filters.dump_raw(system)

    system.setup()

    # show data
    if show_data is not None:
        if len(show_data) == 0:
            show_data = sorted(system.devman.devices)

        for mdl in show_data:
            if system.__dict__[mdl].n == 0:
                continue
            logger.info('Model <{}> data in system base'.format(mdl))
            logger.info(system.__dict__[mdl].data_to_df(sysbase=True).to_string())
        logger.info('')

    if routine is None or exit_now is True:
        logger.info('No routine provided. Set argument `routine` to continue')
    else:
        if isinstance(routine, str):
            routine = [routine]

        # run power flow first
        if 'pflow' in routine:
            routine.remove('pflow')

        system.pflow.run()
        system.tds.init()
        system.report.write(content='powerflow')

        # run the rest of the routines
        for r in routine:
            system.__dict__[r.lower()].run()

    # Disable profiler and output results
    if profile:
        pr.disable()

        if system.files.no_output:
            s = io.StringIO()
            nlines = 40
            ps = pstats.Stats(pr, stream=sys.stdout).sort_stats('cumtime')
            ps.print_stats(nlines)
            logger.info(s.getvalue())
            s.close()
        else:
            s = open(system.files.prof, 'w')
            nlines = 999
            ps = pstats.Stats(pr, stream=s).sort_stats('cumtime')
            ps.print_stats(nlines)
            s.close()
            logger.info('cProfile results for job{:s} written.'.format(
                ' ' + str(pid) if pid >= 0 else ''))

    if pid >= 0:
        _, s = elapsed(t0)
        msg_finish = 'Process {:d} finished in {:s}.'.format(pid, s)
        logger.info(msg_finish)
        print(msg_finish)

    return system


def run_stock(rpath, routine=None, profile=False, dump_raw=False, pid=-1, show_data=None, exit_now=False,
              **kwargs):
    """Run a stock case distributed with andes"""
    case_path = utils.stock_case.get_stock_case(rpath)

    return run(case_path, routine=routine, profile=profile, dump_raw=dump_raw, pid=pid,
               show_data=show_data, exit_now=exit_now, **kwargs)


if __name__ == '__main__':
    main()



import configparser
import logging

from ..utils.cached import cached
from ..utils.tab import Tab

logger = logging.getLogger(__name__)


class ConfigBase(object):
    """
    Base class for routine configurations. This class provides functions for storing,
    accessing, exporting configurations for Andes routines.
    """
    def __init__(self, conf=None, **kwargs):

        if conf is not None:
            self.load_config(conf)

        self.check()

    def get_value(self, option):
        """
        Return the value of the given option

        Parameters
        ----------
        option: str
            option name to retrieve value

        Returns
        -------
        str
            the current value for the option
        """

        return self.__dict__[option]

    def get_alt(self, option):
        """
        Return the available alternative values for an option

        Parameters
        ----------
        option: str
            Name of the option to query

        Returns
        -------
        str
            A comma-separated string containing all the acceptable alternative values.
            If `option` is not valid or no alternative value is on file,
            return an empty string.

        """
        if not hasattr(self, option):
            logger.error('Option <{}> is not valid'.format(option))
            return ''

        alt = option + '_alt'
        if not hasattr(self, alt):
            return ''
        return ','.join(self.__dict__[alt])

    def doc(self, export='plain'):
        """
        Dump help document for setting classes

        Parameters
        ----------
        export : {'plain'}, optional
            The format of the exported config help docs.

        Returns
        -------
        str
            A string containing the formattted config help docs

        """
        rows = []
        title = '<{:s}> config options'.format(self.__class__.__name__)
        table = Tab(export=export, title=title)

        for opt in sorted(self.config_descr):
            if hasattr(self, opt):
                c1 = opt
                c2 = self.config_descr[opt]
                c3 = self.__dict__.get(opt, '')
                c4 = self.get_alt(opt)
                rows.append([c1, c2, c3, c4])
            else:
                logger.error('Setting {:s} has no {:s} option. Correct in config_descr.'.
                             format(self.__class__.__name__, opt))

        table.add_rows(rows, header=False)
        table.header(['Option', 'Description', 'Value', 'Alt.'])

        return table.draw()

    def dump_conf(self, conf=None):
        """
        Dump routine configuration to an rc config file

        Parameters
        ----------
        conf : configparser.ConfigParser, optional
            A config parser object to which this class will be exported. A new object
            will be created if `conf` is `None`.

        Returns
        -------
        configparser.ConfigParser
            A config parser object with the values appended
        """
        if conf is None:
            conf = configparser.ConfigParser()

        tab = self.__class__.__name__

        conf[tab] = {}

        for key, val in self.__dict__.items():
            if key.endswith('_alt'):
                continue
            conf[tab][key] = str(val)

        return conf

    def load_config(self, conf):
        """
        Load configurations from an rc file

        Parameters
        ----------
        rc: str
            path to the rc file

        Returns
        -------
        None
        """
        section = self.__class__.__name__

        if section not in conf.sections():
            logger.debug('Config section {} not in rc file'.format(
                self.__class__.__name__))
            return

        for key in conf[section].keys():
            if not hasattr(self, key):
                logger.debug('Config key {}.{} skipped'.format(section, key))
                continue

            val = conf[section].get(key)

            try:
                val = conf[section].getfloat(key)
            except ValueError:
                try:
                    val = conf[section].getboolean(key)
                except ValueError:
                    pass

            self.__dict__.update({key: val})

        self.check()

    def check(self):
        """
        Check for consistency

        Returns
        -------
        bool
            True for pass, False for fail
        """
        # TODO: check key in _alt
        return True

    @cached
    def config_descr(self):
        descriptions = {}
        return descriptions



class RoutineBase(object):
    """
    Base class for Routines
    """

    def run(self, **kwargs):
        """
        Entry function for power flow routine

        Returns
        -------
        bool
            Success flag
        """

        raise NotImplementedError('Must be overloaded by routines')

    def reset(self):
        """
        Reset internal states of the routine

        Returns
        -------
        None
        """
        raise NotImplementedError('Must be overloaded by routines')

    def report(self):
        """
        Format report from the results

        Returns
        -------
        str
            Output string
        """
        raise NotImplementedError('Must be overloaded by routines')

    def pre(self):
        """
        Pre-check for routine

        Returns
        -------
        bool
            Success flag
        """
        pass

    def post(self):
        """
        Post processing for routine

        Returns
        -------
        bool
            Success flag
        """
        pass



"""PSS/E file parser"""

import logging
import re

from ..consts import deg2rad
from ..utils.math import to_number

logger = logging.getLogger(__name__)


def testlines(fid):
    """Check the raw file for frequency base"""
    first = fid.readline()
    first = first.strip().split('/')
    first = first[0].split(',')
    if float(first[5]) == 50.0 or float(first[5]) == 60.0:
        return True
    else:
        return False


def get_nol(b, mdata):
    """
    Return the number of lines based on data
    """
    nol = [1, 1, 1, 1, 1, 4, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0]

    if b == 5:  # for transformer
        if mdata[0][2] == 0:  # two-winding transformer
            return 4
        else:  # three-widning transf
            return 5

    return nol[b]


def read(file, system):
    """read PSS/E RAW file v32 format"""

    blocks = [
        'bus', 'load', 'fshunt', 'gen', 'branch', 'transf', 'area',
        'twotermdc', 'vscdc', 'impedcorr', 'mtdc', 'msline', 'zone',
        'interarea', 'owner', 'facts', 'swshunt', 'gne', 'Q'
    ]
    nol = [1, 1, 1, 1, 1, 4, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0]
    rawd = re.compile(r'rawd\d\d')

    retval = True
    version = 0
    b = 0  # current block index
    raw = {}
    for item in blocks:
        raw[item] = []

    data = []
    mdata = []  # multi-line data
    mline = 0  # line counter for multi-line models

    # parse file into raw with to_number conversions
    fid = open(file, 'r')
    for num, line in enumerate(fid.readlines()):
        line = line.strip()
        if num == 0:  # get basemva and frequency
            data = line.split('/')[0]
            data = data.split(',')

            mva = float(data[1])
            system.config.mva = mva
            system.config.freq = float(data[5])

            # get raw file version
            version = int(data[2])
            if not version:
                version = int(rawd.search(line).group(0).strip('rawd'))
            logger.debug('PSSE raw version {} detected'.format(version))

            if version < 32 or version > 33:
                logger.warning(
                    'RAW file version is not 32 or 33. Error may occur.')
            continue

        elif num == 1:  # store the case info line
            logger.info(line)
            continue
        elif num == 2:
            continue
        elif num >= 3:
            if line[0:2] == '0 ' or line[0:3] == ' 0 ':  # end of block
                b += 1
                continue
            elif line[0] == 'Q':  # end of file
                break
            data = line.split(',')

        data = [to_number(item) for item in data]
        mdata.append(data)
        mline += 1
        nol = get_nol(b, mdata)

        if mline >= nol:
            if nol == 1:
                mdata = mdata[0]
            raw[blocks[b]].append(mdata)
            mdata = []
            mline = 0
    fid.close()

    # add device elements to system
    sw = {}  # idx:a0
    max_bus = []
    for data in raw['bus']:
        # version 32:
        #   0,   1,      2,     3,    4,   5,  6,   7,  8
        #   ID, NAME, BasekV, Type, Area Zone Owner Vm, Va
        #
        idx = data[0]
        max_bus.append(idx)
        ty = data[3]
        a0 = data[8] * deg2rad

        if ty == 3:
            sw[idx] = a0
        param = {
            'idx': idx,
            'name': data[1],
            'Vn': data[2],
            'voltage': data[7],
            'angle': a0,
            'area': data[4],
            'zone': data[5],
            'owner': data[6],
        }
        system.Bus.elem_add(**param)
    max_bus = max(max_bus)
    for data in raw['load']:
        # version 32:
        #  0,  1,      2,    3,    4,    5,    6,      7,   8,  9, 10,   11
        # Bus, Id, Status, Area, Zone, PL(MW), QL (MW), IP, IQ, YP, YQ, OWNER
        #
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        voltage = system.Bus.get_field('voltage', bus)
        param = {
            'bus': bus,
            'Vn': vn,
            'Sn': mva,
            'p': (data[5] + data[7] * voltage + data[9] * voltage**2) / mva,
            'q': (data[6] + data[8] * voltage - data[10] * voltage**2) / mva,
            'owner': data[11],
        }
        system.PQ.elem_add(**param)

    for data in raw['fshunt']:
        #
        # 0,    1,      2,      3,      4
        # Bus, name, Status, g (MW), b (Mvar)
        #
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        param = {
            'bus': bus,
            'Vn': vn,
            'u': data[2],
            'Sn': mva,
            'g': data[3] / mva,
            'b': data[4] / mva,
        }
        system.Shunt.elem_add(**param)

    gen_idx = 0
    for data in raw['gen']:
        #
        #  0, 1, 2, 3, 4, 5, 6, 7,    8,   9,10,11, 12, 13, 14,   15, 16,17,18,19
        #  I,ID,PG,QG,QT,QB,VS,IREG,MBASE,ZR,ZX,RT,XT,GTAP,STAT,RMPCT,PT,PB,O1,F1
        #
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        gen_mva = data[8]  # unused yet
        gen_idx += 1
        status = data[14]
        param = {
            'Sn': gen_mva,
            'Vn': vn,
            'u': status,
            'idx': gen_idx,
            'bus': bus,
            'pg': status * data[2] / mva,
            'qg': status * data[3] / mva,
            'qmax': data[4] / mva,
            'qmin': data[5] / mva,
            'v0': data[6],
            'ra': data[9],  # ra  armature resistance
            'xs': data[10],  # xs synchronous reactance
            'pmax': data[16] / mva,
            'pmin': data[17] / mva,
        }
        if data[0] in sw.keys():
            param.update({
                'a0': sw[data[0]],
            })
            system.SW.elem_add(**param)
        else:
            system.PV.elem_add(**param)

    for data in raw['branch']:
        #
        # I,J,CKT,R,X,B,RATEA,RATEB,RATEC,GI,BI,GJ,BJ,ST,LEN,O1,F1,...,O4,F4
        #
        param = {
            'bus1': data[0],
            'bus2': data[1],
            'r': data[3],
            'x': data[4],
            'b': data[5],
            'rate_a': data[6],
            'Vn': system.Bus.get_field('Vn', data[0]),
            'Vn2': system.Bus.get_field('Vn', data[1]),
        }
        system.Line.elem_add(**param)

    xf_3_count = 1
    for data in raw['transf']:
        if len(data) == 4:
            # """
            # I,J,K,CKT,CW,CZ,CM,MAG1,MAG2,NMETR,'NAME',STAT,O1,F1,...,O4,F4
            # R1-2,X1-2,SBASE1-2
            # WINDV1,NOMV1,ANG1,RATA1,RATB1,RATC1,COD1,CONT1,RMA1,RMI1,VMA1,VMI1,NTP1,TAB1,CR1,CX1
            # WINDV2,NOMV2
            # """

            tap = data[2][0]
            phi = data[2][2]

            if tap == 1 and phi == 0:
                trasf = False
            else:
                trasf = True
            param = {
                'trasf': trasf,
                'bus1': data[0][0],
                'bus2': data[0][1],
                'u': data[0][11],
                'b': data[0][8],
                'r': data[1][0],
                'x': data[1][1],
                'tap': tap,
                'phi': phi,
                'rate_a': data[2][3],
                'Vn': system.Bus.get_field('Vn', data[0][0]),
                'Vn2': system.Bus.get_field('Vn', data[0][1]),
            }
            system.Line.elem_add(**param)
        else:
            # """
            # I, J, K, CKT, CW, CZ, CM, MAG1, MAG2, NMETR, 'NAME', STAT, Ol, Fl,...,o4, F4
            # R1—2, X1—2, SBASE1—2, R2—3, X2—3, SBASE2—3, R3—1, X3—1, SBASE3—1, VMSTAR, ANSTAR
            # WINDV1, NOMV1, ANG1, RATA1, BATB1, RATC1, COD1, CONT1, RMA1, RMI1, VMA1, VMI1, NTP1, TAB1, CR1, CX1
            # WINDV2, NOMV2, ANG2, RATA2, BATB2, RATC2, COD2, CONT2, RMA2, RMI2, VMA2, VMI2, NTP2, TAB2, CR2, CX2
            # WINDV3, NOMV3, ANG3, RATA3, BATB3, RATC3, COD3, CONT3, RMA3, RMI3, VMA3, VMI3, NTP3, TAB3, CR3, CX3
            # """
            param = {
                'idx': max_bus+xf_3_count,
                'name': '_'.join(str(data[0][:3])),
                'Vn': system.Bus.get_field('Vn', data[0][0]),
                'voltage': data[1][-2],
                'angle': data[1][-1]/180*3.14
            }
            system.Bus.elem_add(**param)

            r = []
            x = []
            r.append((data[1][0] + data[1][6] - data[1][3])/2)
            r.append((data[1][3] + data[1][0] - data[1][6])/2)
            r.append((data[1][6] + data[1][3] - data[1][0])/2)
            x.append((data[1][1] + data[1][7] - data[1][4])/2)
            x.append((data[1][4] + data[1][1] - data[1][7])/2)
            x.append((data[1][7] + data[1][4] - data[1][1])/2)
            for i in range(0, 3):
                param = {
                    'trasf': True,
                    'bus1': data[0][i],
                    'bus2': max_bus+xf_3_count,
                    'u': data[0][11],
                    'b': data[0][8],
                    'r': r[i],
                    'x': x[i],
                    'tap': data[2+i][0],
                    'phi': data[2+i][2],
                    'rate_a': data[2+i][3],
                    'Vn': system.Bus.get_field('Vn', data[0][i]),
                    'Vn2': system.Bus.get_field('Vn', data[0][0]),
                }
                system.Line.elem_add(**param)
            xf_3_count += 1

    for data in raw['swshunt']:
        # I, MODSW, ADJM, STAT, VSWHI, VSWLO, SWREM, RMPCT, ’RMIDNT’,
        # BINIT, N1, B1, N2, B2, ... N8, B8
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        param = {
            'bus': bus,
            'Vn': vn,
            'Sn': mva,
            'u': data[3],
            'b': data[9] / mva,
        }
        system.Shunt.elem_add(**param)

    for data in raw['area']:
        # ID, ISW, PDES, PTOL, ARNAME
        param = {
            'idx': data[0],
            'isw': data[1],
            'pdes': data[2],
            'ptol': data[3],
            'name': data[4],
        }
        system.Area.elem_add(**param)

    for data in raw['zone']:
        # """ID, NAME"""
        param = {
            'idx': data[0],
            'name': data[1],
        }
        system.Zone.elem_add(**param)

    return retval


def readadd(file, system):
    """read DYR file"""
    dyr = {}
    data = []
    end = 0
    retval = True
    sep = ','

    fid = open(file, 'r')
    for line in fid.readlines():
        if line.find('/') >= 0:
            line = line.split('/')[0]
            end = 1
        if line.find(',') >= 0:  # mixed comma and space splitter not allowed
            line = [to_number(item.strip()) for item in line.split(sep)]
        else:
            line = [to_number(item.strip()) for item in line.split()]
        if not line:
            end = 0
            continue
        data.extend(line)
        if end == 1:
            field = data[1]
            if field not in dyr.keys():
                dyr[field] = []
            dyr[field].append(data)
            end = 0
            data = []
    fid.close()

    # elem_add device elements to system
    supported = [
        'GENROU',
        'GENCLS',
        'ESST3A',
        'ESDC2A',
        'SEXS',
        'EXST1',
        'ST2CUT',
        'IEEEST',
        'TGOV1',
    ]
    used = list(supported)
    for model in supported:
        if model not in dyr.keys():
            used.remove(model)
            continue
        for data in dyr[model]:
            add_dyn(system, model, data)

    needed = list(dyr.keys())
    for i in supported:
        if i in needed:
            needed.remove(i)

    if len(needed) > 0:
        logger.warning('Models currently unsupported: {}'.format(', '.join(needed)))

    return retval


def add_dyn(system, model, data):
    """helper function to elem_add a device element to system"""
    if model == 'GENCLS':
        bus = data[0]
        data = data[3:]
        if bus in system.PV.bus:
            dev = 'PV'
            gen_idx = system.PV.idx[system.PV.bus.index(bus)]
        elif bus in system.SW.bus:
            dev = 'SW'
            gen_idx = system.SW.idx[system.SW.bus.index(bus)]
        else:
            raise KeyError
        # todo: check xl
        idx_PV = get_idx(system, 'StaticGen', 'bus', bus)
        u = get_param(system, 'StaticGen', 'u', idx_PV)
        param = {
            'bus': bus,
            'gen': gen_idx,
            # 'idx': bus,
            #  use `bus` for `idx`. Only one generator allowed on each bus
            'Sn': system.__dict__[dev].get_field('Sn', gen_idx),
            'Vn': system.__dict__[dev].get_field('Vn', gen_idx),
            'xd1': system.__dict__[dev].get_field('xs', gen_idx),
            'ra': system.__dict__[dev].get_field('ra', gen_idx),
            'M': 2 * data[0],
            'D': data[1],
            'u': u,
        }
        system.Syn2.elem_add(**param)

    elif model == 'GENROU':
        bus = data[0]
        data = data[3:]
        if bus in system.PV.bus:
            dev = 'PV'
            gen_idx = system.PV.idx[system.PV.bus.index(bus)]
        elif bus in system.SW.bus:
            dev = 'SW'
            gen_idx = system.SW.idx[system.SW.bus.index(bus)]
        else:
            raise KeyError
        idx_PV = get_idx(system, 'StaticGen', 'bus', bus)
        u = get_param(system, 'StaticGen', 'u', idx_PV)

        param = {
            'bus': bus,
            'gen': gen_idx,
            # 'idx': bus,
            #  use `bus` for `idx`. Only one generator allowed on each bus
            'Sn': system.__dict__[dev].get_field('Sn', gen_idx),
            'Vn': system.__dict__[dev].get_field('Vn', gen_idx),
            'ra': system.__dict__[dev].get_field('ra', gen_idx),
            'Td10': data[0],
            'Td20': data[1],
            'Tq10': data[2],
            'Tq20': data[3],
            'M': 2 * data[4],
            'D': data[5],
            'xd': data[6],
            'xq': data[7],
            'xd1': data[8],
            'xq1': data[9],
            'xd2': data[10],
            'xq2': data[10],  # xd2 = xq2
            'xl': data[11],
            'u': u,
        }
        system.Syn6a.elem_add(**param)

    elif model == 'ESST3A':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[8],
            'vrmin': data[9],
            'Ka': data[6],
            'Ta': data[7],
            'Tf': data[5],
            'Tr': data[0],
            'Kf': data[5],
            'Ke': 1,
            'Te': 1,
        }
        system.AVR1.elem_add(**param)

    elif model == 'ESDC2A':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[5],
            'vrmin': data[6],
            'Ka': data[1],
            'Ta': data[2],
            'Tf': data[10],
            'Tr': data[0],
            'Kf': data[9],
            'Ke': 1,
            'Te': data[8],
        }
        system.AVR1.elem_add(**param)

    elif model == 'EXST1':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[7],
            'vrmin': data[8],
            'Ka': data[5],
            'Ta': data[6],
            'Kf': data[10],
            'Tf': data[11],
            'Tr': data[0],
            'Te': data[4],
        }
        system.AVR1.elem_add(**param)

    elif model == 'SEXS':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[5],
            'vrmin': data[4],
            'K0': data[2],
            'T2': data[1],
            'T1': data[0],
            'Te': data[3],
        }
        system.AVR3.elem_add(**param)

    elif model == 'IEEEG1':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)

    elif model == 'TGOV1':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'gen': syn,
            'R': data[0],
            'T1': data[4],
            'T2': data[5],
        }
        system.TG2.elem_add(**param)

    elif model == 'ST2CUT':
        bus = data[0]
        data = data[3:]
        Ic1 = data[0]
        Ic2 = data[2]

        data = data[4:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        avr = get_idx(system, 'AVR', 'syn', syn)
        param = {
            'avr': avr,
            'Ic1': Ic1,
            'Ic2': Ic2,
            'K1': data[0],
            'K2': data[1],
            'T1': data[2],
            'T2': data[3],
            'T3': data[4],
            'T4': data[5],
            'T5': data[6],
            'T6': data[7],
            'T7': data[8],
            'T8': data[9],
            'T9': data[10],
            'T10': data[11],
            'lsmax': data[12],
            'lsmin': data[13],
            'vcu': data[14],
            'vcl': data[15],
        }
        system.PSS1.elem_add(**param)

    elif model == 'IEEEST':
        bus = data[0]
        data = data[3:]
        Ic = data[0]

        data = data[2:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        avr = get_idx(system, 'AVR', 'syn', syn)
        param = {
            'avr': avr,
            'Ic': Ic,
            'A1': data[0],
            'A2': data[1],
            'A3': data[2],
            'A4': data[3],
            'A5': data[4],
            'A6': data[5],
            'T1': data[6],
            'T2': data[7],
            'T3': data[8],
            'T4': data[9],
            'T5': data[10],
            'T6': data[11],
            'Ks': data[12],
            'lsmax': data[13],
            'lsmin': data[14],
            'vcu': data[15],
            'vcl': data[16],
        }
        system.PSS2.elem_add(**param)

    else:
        logger.warning('Skipping unsupported model <{}> on bus {}'.format(
            model, data[0]))


def get_idx(system, group, param, fkey):
    ret = None
    for key, item in system.devman.group.items():
        if key != group:
            continue
        for name, dev in item.items():
            int_id = system.__dict__[dev].uid[name]
            if system.__dict__[dev].__dict__[param][int_id] == fkey:
                ret = name
                break
    return ret


def get_param(system, group, param, fkey):
    ret = None
    for key, item in system.devman.group.items():
        if key != group:
            continue
        for name, dev in item.items():
            if name == fkey:
                int_id = system.__dict__[dev].uid[name]
                ret = system.__dict__[dev].__dict__[param][int_id]
    return ret



import logging

logger = logging.getLogger(__name__)

SHOW_PF_CALL = False
SHOW_INT_CALL = False

all_calls = [
    'gcall', 'gycall', 'fcall', 'fxcall', 'init0', 'pflow', 'windup', 'jac0',
    'init1', 'shunt', 'series', 'flows', 'connection', 'times', 'stagen',
    'dyngen', 'gmcall', 'fmcall', 'dcseries', 'opf', 'obj'
]


class Call(object):
    """ Equation call mamager class for andes routines"""

    def __init__(self, system):
        self.system = system
        self.ndevice = 0
        self.devices = []

        for item in all_calls:
            self.__dict__[item] = []

    def setup(self):
        """
        setup the call list after case file is parsed and jit models are loaded
        """
        self.devices = self.system.devman.devices
        self.ndevice = len(self.devices)

        self.build_vec()

    def build_vec(self):
        """build call validity vector for each device"""
        for item in all_calls:
            self.__dict__[item] = []

        for dev in self.devices:
            for item in all_calls:
                if self.system.__dict__[dev].n == 0:
                    val = False
                else:
                    val = self.system.__dict__[dev].calls.get(item, False)
                self.__dict__[item].append(val)

    def fdpf(self):
        system = self.system
        system.dae.init_g()
        for device, pflow, gcall in zip(self.devices, self.pflow):
            if pflow and gcall:
                system.__dict__[device].gcall(system.dae)
        system.dae.reset_small()

    def pfload(self):
        system = self.system
        system.dae.init_g()
        for device, gcall, pflow, shunt, stagen in zip(self.devices, self.gcall, self.pflow, self.shunt,
                                                       self.stagen):
            if gcall and pflow and shunt and not stagen:
                system.__dict__[device].gcall(system.dae)
        system.dae.reset_small_g()

    def pfgen(self):
        system = self.system
        system.dae.init_g()
        for device, gcall, pflow, shunt, series, stagen in zip(self.devices, self.gcall, self.pflow, self.shunt,
                                                               self.series, self.stagen):
            if gcall and pflow and (shunt or series) and not stagen:
                system.__dict__[device].gcall(system.dae)
        system.dae.reset_small_g()

    def bus_injection(self):
        system = self.system
        for device, series in zip(self.devices, self.series):
            if series:
                system.__dict__[device].gcall(system.dae)
        system.dae.reset_small_g()
        self.gisland()

    def gisland(self):
        self.system.Bus.gisland(self.system.dae)

    def gyisland(self):
        self.system.Bus.gyisland(self.system.dae)

    def seriesflow(self):
        system = self.system
        for device, pflow, series in zip(self.devices, self.pflow, self.series):
            if pflow and series:
                system.__dict__[device].seriesflow(system.dae)

    def int_fg(self):
        system = self.system
        system.dae.init_fg(resetz=False)
        for device, gcall in zip(self.devices, self.gcall):
            if gcall:
                system.__dict__[device].gcall(system.dae)
        system.dae.reset_small_g()
        self.gisland()

        for device, fcall in zip(self.devices, self.fcall):
            if fcall:
                system.__dict__[device].fcall(system.dae)
        system.dae.reset_small_f()

    def int_fxgy(self):
        system = self.system
        # rebuilt constant elements in jacobian if needed
        if system.dae.factorize:
            system.dae.init_jac0()
            for device, jac0 in zip(self.devices, self.jac0):
                if jac0:
                    system.__dict__[device].jac0(system.dae)
            system.dae.temp_to_spmatrix('jac0')

        system.dae.setup_FxGy()

        self.gyisland()
        for device, gycall in zip(self.devices, self.gycall):
            if gycall:
                system.__dict__[device].gycall(system.dae)

        for device, fxcall in zip(self.devices, self.fxcall):
            if fxcall:
                system.__dict__[device].fxcall(system.dae)

        system.dae.temp_to_spmatrix('jac')

    def int(self):
        self.int_fg()
        self.int_fxgy()



import importlib
import logging

from cvxopt import matrix, spmatrix, sparse, spdiag, printing  # NOQA
from cvxopt import mul, div  # NOQA

from .base import ModelBase
from ..consts import Gy  # NOQA
from ..consts import deg2rad
from ..utils.math import polar, conj

logger = logging.getLogger(__name__)

pd = None
printing.options['dformat'] = '%.3f'
printing.options['width'] = -1


class Line(ModelBase):
    """AC transmission line lumped model"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'Series'
        self._name = 'Line'
        self._data.update({
            'r': 0.0,
            'x': 1e-6,
            'b': 0.0,
            'g': 0.0,
            'b1': 0.0,
            'g1': 0.0,
            'b2': 0.0,
            'g2': 0.0,
            'bus1': None,
            'bus2': None,
            'Vn2': 110.0,
            'xcoord': None,
            'ycoord': None,
            'trasf': False,
            'tap': 1.0,
            'phi': 0,
            'fn': 60,
            'owner': 0,
            'rate_a': 0,
        })
        self._units.update({
            'r': 'pu',
            'x': 'pu',
            'b': 'pu',
            'g': 'pu',
            'b1': 'pu',
            'g1': 'pu',
            'b2': 'pu',
            'g2': 'pu',
            'bus1': 'na',
            'bus2': 'na',
            'Vn2': 'kV',
            'xcoord': 'deg',
            'ycoord': 'deg',
            'trasf': 'na',
            'tap': 'na',
            'phi': 'deg',
            'fn': 'Hz',
            'owner': 'na',
            'rate_a': 'pu',
        })
        self._descr.update({
            'r': 'connection line resistance',
            'x': 'connection line reactance',
            'g': 'shared shunt conductance',
            'b': 'shared shunt susceptance',
            'g1': 'from-side conductance',
            'b1': 'from-side susceptance',
            'g2': 'to-side conductance',
            'b2': 'to-side susceptance',
            'bus1': 'idx of from bus',
            'bus2': 'idx of to bus',
            'Vn2': 'rated voltage of bus2',
            'xcoord': 'x coordinates',
            'ycoord': 'y coordinates',
            'trasf': 'transformer branch flag',
            'tap': 'transformer branch tap ratio',
            'phi': 'transformer branch phase shift in rad',
            'fn': 'rated frequency',
            'owner': 'owner code',
            'rate_a': 'phase a power flow limit',
        })
        self._params.extend([
            'r', 'x', 'b', 'g', 'b1', 'g1', 'b2', 'g2', 'tap', 'phi', 'fn',
            'rate_a'
        ])
        self._service.extend([
            'a', 'v', 'a1', 'a2', 'S1', 'S2', 'nb', 'gy_store', 'y1', 'y2',
            'y12', 'm', 'm2', 'mconj'
        ])
        self.calls.update({
            'gcall': True,
            'gycall': True,
            'init0': True,
            'pflow': True,
            'series': True,
            'flows': True
        })

        self._ac = {'bus1': ['a1', 'v1'], 'bus2': ['a2', 'v2']}

        self._config['is_series'] = True

        self.rebuild = True
        self.Y = []
        self.C = []
        self.Bp = []
        self.Bpp = []
        self._init()

    def setup(self):
        self._param_to_matrix()

        self.nb = int(self.system.Bus.n)
        self.system.config.nseries += self.n

        self.r += 1e-10
        self.b += 1e-10

        self.g1 += 0.5 * self.g
        self.b1 += 0.5 * self.b
        self.g2 += 0.5 * self.g
        self.b2 += 0.5 * self.b

    def build_y(self):
        """Build transmission line admittance matrix into self.Y"""
        if not self.n:
            return
        self.y1 = mul(self.u, self.g1 + self.b1 * 1j)
        self.y2 = mul(self.u, self.g2 + self.b2 * 1j)
        self.y12 = div(self.u, self.r + self.x * 1j)
        self.m = polar(self.tap, self.phi * deg2rad)
        self.m2 = abs(self.m)**2
        self.mconj = conj(self.m)

        # build self and mutual admittances into Y
        self.Y = spmatrix(
            div(self.y12 + self.y1, self.m2), self.a1, self.a1,
            (self.nb, self.nb), 'z')
        self.Y -= spmatrix(
            div(self.y12, self.mconj), self.a1, self.a2, (self.nb, self.nb),
            'z')
        self.Y -= spmatrix(
            div(self.y12, self.m), self.a2, self.a1, (self.nb, self.nb), 'z')
        self.Y += spmatrix(self.y12 + self.y2, self.a2, self.a2,
                           (self.nb, self.nb), 'z')

        # avoid singularity
        # for item in range(self.nb):
        #     if abs(self.Y[item, item]) == 0:
        #         self.Y[item, item] = 1e-6 + 0j

    def build_b(self):
        """build Bp and Bpp for fast decoupled method"""
        if not self.n:
            return
        method = self.system.pflow.config.method.lower()

        # Build B prime matrix
        y1 = mul(
            self.u, self.g1
        )  # y1 neglects line charging shunt, and g1 is usually 0 in HV lines
        y2 = mul(
            self.u, self.g2
        )  # y2 neglects line charging shunt, and g2 is usually 0 in HV lines
        m = polar(1.0, self.phi * deg2rad)  # neglected tap ratio
        self.mconj = conj(m)
        m2 = matrix(1.0, (self.n, 1), 'z')
        if method in ('fdxb', 'dcpf'):
            # neglect line resistance in Bp in XB method
            y12 = div(self.u, self.x * 1j)
        else:
            y12 = div(self.u, self.r + self.x * 1j)
        self.Bp = spmatrix(
            div(y12 + y1, m2), self.a1, self.a1, (self.nb, self.nb), 'z')
        self.Bp -= spmatrix(
            div(y12, conj(m)), self.a1, self.a2, (self.nb, self.nb), 'z')
        self.Bp -= spmatrix(
            div(y12, m), self.a2, self.a1, (self.nb, self.nb), 'z')
        self.Bp += spmatrix(y12 + y2, self.a2, self.a2, (self.nb, self.nb),
                            'z')
        self.Bp = self.Bp.imag()

        # Build B double prime matrix
        y1 = mul(
            self.u, self.g1 + self.b1 * 1j
        )  # y1 neglected line charging shunt, and g1 is usually 0 in HV lines
        y2 = mul(
            self.u, self.g2 + self.b2 * 1j
        )  # y2 neglected line charging shunt, and g2 is usually 0 in HV lines
        m = self.tap + 0j  # neglected phase shifter
        m2 = abs(m)**2 + 0j

        if method in ('fdbx', 'fdpf', 'dcpf'):
            # neglect line resistance in Bpp in BX method
            y12 = div(self.u, self.x * 1j)
        else:
            y12 = div(self.u, self.r + self.x * 1j)
        self.Bpp = spmatrix(
            div(y12 + y1, m2), self.a1, self.a1, (self.nb, self.nb), 'z')
        self.Bpp -= spmatrix(
            div(y12, conj(m)), self.a1, self.a2, (self.nb, self.nb), 'z')
        self.Bpp -= spmatrix(
            div(y12, m), self.a2, self.a1, (self.nb, self.nb), 'z')
        self.Bpp += spmatrix(y12 + y2, self.a2, self.a2, (self.nb, self.nb),
                             'z')
        self.Bpp = self.Bpp.imag()

        for item in range(self.nb):
            if abs(self.Bp[item, item]) == 0:
                self.Bp[item, item] = 1e-6 + 0j
            if abs(self.Bpp[item, item]) == 0:
                self.Bpp[item, item] = 1e-6 + 0j

    def incidence(self):
        """Build incidence matrix into self.C"""
        self.C = \
            spmatrix(self.u, range(self.n), self.a1, (self.n, self.nb), 'd') -\
            spmatrix(self.u, range(self.n), self.a2, (self.n, self.nb), 'd')

    def connectivity(self, bus):
        """check connectivity of network using Goderya's algorithm"""
        if not self.n:
            return
        n = self.nb
        fr = self.a1
        to = self.a2
        os = [0] * self.n

        # find islanded buses
        diag = list(
            matrix(
                spmatrix(self.u, to, os, (n, 1), 'd') +
                spmatrix(self.u, fr, os, (n, 1), 'd')))
        nib = bus.n_islanded_buses = diag.count(0)
        bus.islanded_buses = []
        for idx in range(n):
            if diag[idx] == 0:
                bus.islanded_buses.append(idx)

        # find islanded areas
        temp = spmatrix(
            list(self.u) * 4, fr + to + fr + to, to + fr + fr + to, (n, n),
            'd')
        cons = temp[0, :]
        nelm = len(cons.J)
        conn = spmatrix([], [], [], (1, n), 'd')
        bus.island_sets = []
        idx = islands = 0
        enum = 0

        while 1:
            while 1:
                cons = cons * temp
                cons = sparse(cons)  # remove zero values
                new_nelm = len(cons.J)
                if new_nelm == nelm:
                    break
                nelm = new_nelm
            if len(cons.J) == n:  # all buses are interconnected
                return
            bus.island_sets.append(list(cons.J))
            conn += cons
            islands += 1
            nconn = len(conn.J)
            if nconn >= (n - nib):
                bus.island_sets = [i for i in bus.island_sets if i != []]
                break

            for element in conn.J[idx:]:
                if not diag[idx]:
                    enum += 1  # skip islanded buses
                if element <= enum:
                    idx += 1
                    enum += 1
                else:
                    break

            cons = temp[enum, :]

    def init0(self, dae):
        self.copy_data_ext('Bus', 'a', dest='a', idx=None, astype=list)
        self.copy_data_ext('Bus', 'v', dest='v', idx=None, astype=list)

        method = self.system.pflow.config.method.lower()
        self.build_y()
        self.incidence()
        if method in ('fdpf', 'fdbx', 'fdxb', 'dcpf'):
            self.build_b()

    def gcall(self, dae):
        if self.system.pflow.config.method == 'DCPF':
            return
        if self.rebuild:
            self.build_y()
            self.rebuild = False
        vc = polar(dae.y[self.v], dae.y[self.a])
        Ic = self.Y * vc
        S = mul(vc, conj(Ic))
        dae.g[self.a] += S.real()
        dae.g[self.v] += S.imag()

    def gycall(self, dae):
        gy = self.build_gy(dae)
        dae.add_jac(Gy, gy.V, gy.I, gy.J)

    def build_gy(self, dae):
        """Build line Jacobian matrix"""
        if not self.n:
            idx = range(dae.m)
            dae.set_jac(Gy, 1e-6, idx, idx)
            return

        Vn = polar(1.0, dae.y[self.a])
        Vc = mul(dae.y[self.v], Vn)
        Ic = self.Y * Vc

        diagVn = spdiag(Vn)
        diagVc = spdiag(Vc)
        diagIc = spdiag(Ic)

        dS = self.Y * diagVn
        dS = diagVc * conj(dS)
        dS += conj(diagIc) * diagVn

        dR = diagIc
        dR -= self.Y * diagVc
        dR = diagVc.H.T * dR

        self.gy_store = sparse([[dR.imag(), dR.real()], [dS.real(),
                                                         dS.imag()]])
        return self.gy_store

    def seriesflow(self, dae):
        """
        Compute the flow through the line after solving PF.

        Compute terminal injections, line losses
        """

        # Vm = dae.y[self.v]
        # Va = dae.y[self.a]
        # V1 = polar(Vm[self.a1], Va[self.a1])
        # V2 = polar(Vm[self.a2], Va[self.a2])

        I1 = mul(self.v1, div(self.y12 + self.y1, self.m2)) - mul(self.v2, div(self.y12, self.mconj))
        I2 = mul(self.v2, self.y12 + self.y2) - mul(self.v1, div(self.y12, self.m))

        self.I1_real = I1.real()
        self.I1_imag = I1.imag()
        self.I2_real = I2.real()
        self.I2_imag = I2.imag()

        self.S1 = mul(self.v1, conj(I1))
        self.S2 = mul(self.v2, conj(I2))

        self.P1 = self.S1.real()
        self.P2 = self.S2.real()
        self.Q1 = self.S1.imag()
        self.Q2 = self.S2.imag()

        self.chg1 = mul(self.g1 + 1j * self.b1, div(self.v1**2, self.m2))
        self.chg2 = mul(self.g2 + 1j * self.b2, self.v2**2)

        self.Pchg1 = self.chg1.real()
        self.Pchg2 = self.chg2.real()

        self.Qchg1 = self.chg1.imag()
        self.Qchg2 = self.chg2.imag()

        self._line_flows = matrix([self.P1, self.P2, self.Q1, self.Q2,
                                   self.I1_real, self.I1_imag,
                                   self.I2_real, self.I2_imag])

    @property
    def v1(self):
        """Return voltage phasors at the "from buses" (bus1)"""
        Vm = self.system.dae.y[self.v]
        Va = self.system.dae.y[self.a]
        return polar(Vm[self.a1], Va[self.a1])

    @property
    def v2(self):
        """Return voltage phasors at the "to buses" (bus2)"""
        Vm = self.system.dae.y[self.v]
        Va = self.system.dae.y[self.a]
        return polar(Vm[self.a2], Va[self.a2])

    def switch(self, idx, u):
        """switch the status of Line idx"""
        self.u[self.uid[idx]] = u
        self.rebuild = True
        self.system.dae.factorize = True
        logger.debug('<Line> Status switch to {} on idx {}.'.format(u, idx))

    def build_name_from_bus(self):
        """Rebuild line names from bus names"""
        pass

    def _varname_flow(self):
        """Build variable names for Pij, Pji, Qij, Qji, Sij, Sji"""
        if not self.n:
            return

        mpq = self.system.dae.m + 2 * self.system.Bus.n
        nl = self.n

        # Pij
        xy_idx = range(mpq, mpq + nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Pij',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='P_{ij}',
            element_name=self.name)

        # Pji
        xy_idx = range(mpq + nl, mpq + 2 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Pji',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='P_{ji}',
            element_name=self.name)

        # Qij
        xy_idx = range(mpq + 2 * nl, mpq + 3 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Qij',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='Q_{ij}',
            element_name=self.name)

        # Qji
        xy_idx = range(mpq + 3 * nl, mpq + 4 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Qji',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='Q_{ji}',
            element_name=self.name)

        # Iij Real
        xy_idx = range(mpq + 4 * nl, mpq + 5 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Iij_Re',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='\\Re(I_{ij})',
            element_name=self.name)

        # Iij Imag
        xy_idx = range(mpq + 5 * nl, mpq + 6 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Iij_Im',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='\\Im(I_{ij})',
            element_name=self.name)

        # Iji Real
        xy_idx = range(mpq + 6 * nl, mpq + 7 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Iji_Re',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='\\Re(I_{ji})',
            element_name=self.name)

        # Iji Imag
        xy_idx = range(mpq + 7 * nl, mpq + 8 * nl)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Iji_Im',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='\\Im(I_{ji})',
            element_name=self.name)

    def get_flow_by_idx(self, idx, bus):
        """Return seriesflow based on the external idx on the `bus` side"""
        P, Q = [], []
        if not isinstance(idx, list):
            idx = [idx]
        if not isinstance(bus, list):
            bus = [bus]

        for line_idx, bus_idx in zip(idx, bus):
            line_int = self.uid[line_idx]
            if bus_idx == self.bus1[line_int]:
                P.append(self.P1[line_int])
                Q.append(self.Q1[line_int])
            elif bus_idx == self.bus2[line_int]:
                P.append(self.P2[line_int])
                Q.append(self.Q2[line_int])
        return matrix(P), matrix(Q)

    def leaf_bus(self, df=False):
        """
        Return leaf bus idx, line idx, and the line foreign key

        Returns
        -------
        (list, list, list) or DataFrame

        """
        # leafs - leaf bus idx
        # lines - line idx
        # fkey - the foreign key of Line, in 'bus1' or 'bus2', linking the bus

        leafs, lines, fkeys = list(), list(), list()

        # convert to unique, ordered list
        buses = sorted(list(set(self.bus1 + self.bus2)))

        links = self.link_bus(buses)

        for bus, link in zip(buses, links):
            line = link[0]
            fkey = link[1]
            if line is None:
                continue
            if len(line) == 1:
                leafs.append(bus)
                lines.extend(line)
                fkeys.extend(fkey)

        # output formatting
        if df is False:
            return leafs, lines, fkeys
        else:
            _data = {'Bus idx': leafs, 'Line idx': lines, 'fkey': fkeys}
            if globals()['pd'] is None:
                globals()['pd'] = importlib.import_module('pandas')
            return pd.DataFrame(data=_data)



from cvxopt import matrix, spmatrix, mul, div  # NOQA

from .base import ModelBase
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA


class Node(ModelBase):
    """ DC node class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'Topology'
        self._name = 'Node'
        self.param_remove('Sn')
        self.param_remove('Vn')
        self._params.extend([
            'Vdcn',
            'Idcn',
            'voltage',
        ])
        self._descr.update({
            'Vdcn': 'DC voltage rating',
            'Idcn': 'DC current rating',
            'voltage': 'Initial nodal voltage guess',
            'area': 'Area code',
            'region': 'Region code',
            'xcoord': 'x coordinate',
            'ycoord': 'y coordinate'
        })
        self._data.update({
            'Vdcn': 100.0,
            'Idcn': 10.0,
            'area': 0,
            'region': 0,
            'voltage': 1.0,
            'xcoord': None,
            'ycoord': None,
        })
        self._units.update({
            'Vdcn': 'kV',
            'Idcn': 'kA',
            'area': 'na',
            'region': 'na',
            'voltage': 'pu',
            'xcoord': 'deg',
            'ycoord': 'deg',
        })
        self.calls.update({
            'init0': True,
            'pflow': True,
            'jac0': True,
        })
        self._mandatory = ['Vdcn']
        self._zeros = ['Vdcn', 'Idcn']
        # self.v = list()
        self._algebs.extend(['v'])
        self._fnamey.extend(['V_{dc}'])
        self._init()

    def elem_add(self, idx=None, name=None, **kwargs):
        super().elem_add(idx, name, **kwargs)

    def _varname(self):
        if not self.n:
            return
        self.system.varname.append(
            listname='unamey',
            xy_idx=self.v,
            var_name='Vdc',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=self.v,
            var_name='V_{dc}',
            element_name=self.name)

    def setup(self):
        self._param_to_matrix()

    def init0(self, dae):
        dae.y[self.v] = matrix(self.voltage, (self.n, 1), 'd')

    def jac0(self, dae):
        if self.n == 0:
            return
        dae.add_jac(Gy0, -1e-6, self.v, self.v)


class DCBase(ModelBase):
    """Two-terminal DC device base"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'DCBasic'
        self._params.remove('Sn')
        self._params.remove('Vn')
        self._data.update({
            'node1': None,
            'node2': None,
            'Vdcn': 100.0,
            'Idcn': 10.0,
        })
        self._params.extend(['Vdcn', 'Idcn'])
        self._descr.update({
            'Vdcn': 'DC voltage rating',
            'Idcn': 'DC current rating',
            'node1': 'DC node 1 idx',
            'node2': 'DC node 2 idx',
        })
        self._units.update({
            'Vdcn': 'kV',
            'Idcn': 'A',
        })
        self._dc = {
            'node1': 'v1',
            'node2': 'v2',
        }
        self._mandatory.extend(['node1', 'node2', 'Vdcn'])

    @property
    def v12(self):
        return self.system.dae.y[self.v1] - self.system.dae.y[self.v2]


class R(DCBase):
    """DC Resistence line class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'R'
        self._params.extend(['R'])
        self._data.update({
            'R': 0.01,
        })
        self._r.extend(['R'])
        self.calls.update({
            'pflow': True,
            'gcall': True,
            'jac0': True,
        })
        self._algebs.extend(['Idc'])
        self._fnamey = ['I_{dc}']
        self._init()

    def gcall(self, dae):
        dae.g[self.Idc] = div(self.v12, self.R) + dae.y[self.Idc]
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def jac0(self, dae):
        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)
        dae.add_jac(Gy0, div(self.u, self.R), self.Idc, self.v1)
        dae.add_jac(Gy0, -div(self.u, self.R), self.Idc, self.v2)
        dae.add_jac(Gy0, self.u - 1e-6, self.Idc, self.Idc)


class L(DCBase):
    """Pure inductive line"""

    def __init__(self, system, name):
        super(L, self).__init__(system, name)
        self._name = 'L'
        self._data.update({'L': 0.001})
        self._params.extend(['L'])
        self._r.extend(['L'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['IL'])
        self._fnamex.extend(['I_L'])
        self._service.extend(['iL'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._init()

    def servcall(self, dae):
        self.iL = div(self.u, self.L)

    def init0(self, dae):
        self.servcall(dae)

    def gcall(self, dae):
        dae.g[self.Idc] = dae.x[self.IL] + dae.y[self.Idc]
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        dae.f[self.IL] = mul(self.v12, self.iL)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.IL)
        dae.add_jac(Gy0, self.u, self.Idc, self.Idc)
        dae.add_jac(Fy0, self.iL, self.IL, self.v1)
        dae.add_jac(Fy0, -self.iL, self.IL, self.v2)
        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)


class C(DCBase):
    """Pure capacitive line"""

    def __init__(self, system, name):
        super(C, self).__init__(system, name)
        self._name = 'C'
        self._data.update({'C': 0.001})
        self._params.extend(['C'])
        self._g.extend(['C'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['vC'])
        self._fnamex.extend(['vC'])
        self._service.extend(['iC'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._init()

    def servcall(self, dae):
        self.iC = div(self.u, self.C)

    def init0(self, dae):
        self.servcall(dae)

    def gcall(self, dae):
        dae.g[self.Idc] = dae.x[self.vC] - self.v12
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        dae.f[self.vC] = -mul(dae.y[self.Idc], self.iC)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.vC)
        dae.add_jac(Gy0, -self.u, self.Idc, self.v1)
        dae.add_jac(Gy0, self.u, self.Idc, self.v2)
        dae.add_jac(Gy0, 1e-6, self.Idc, self.Idc)

        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)

        dae.add_jac(Fy0, -self.iC, self.vC, self.Idc)


class RLs(DCBase):
    """DC Resistive and Inductive line"""

    def __init__(self, system, name):
        super(RLs, self).__init__(system, name)
        self._name = 'RLs'
        self._params.extend(['R', 'L'])
        self._data.update({
            'R': 0.01,
            'L': 0.001,
        })
        self._params.extend(['R', 'L'])
        self._r.extend(['R', 'L'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['IL'])
        self._fnamex.extend(['I_L'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._service.extend(['iR', 'iL'])
        self._init()

    def servcall(self, dae):
        self.iR = div(self.u, self.R)
        self.iL = div(self.u, self.L)

    def init0(self, dae):
        self.servcall(dae)
        dae.x[self.IL] = mul(self.v12, self.iR)
        dae.y[self.Idc] = -dae.x[self.IL]

    def gcall(self, dae):
        dae.g[self.Idc] = mul(self.u, dae.x[self.IL] + dae.y[self.Idc])
        dae.g -= spmatrix(
            mul(self.u, dae.y[self.Idc]), self.v1, [0] * self.n, (dae.m, 1),
            'd')
        dae.g += spmatrix(
            mul(self.u, dae.y[self.Idc]), self.v2, [0] * self.n, (dae.m, 1),
            'd')

    def fcall(self, dae):
        dae.f[self.IL] = mul(self.v12 - mul(self.R, dae.x[self.IL], self.u),
                             self.iL)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.IL)
        dae.add_jac(Gy0, self.u + 1e-6, self.Idc, self.Idc)
        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)
        dae.add_jac(Fx0, -mul(self.R, self.iL, self.u) + 1e-6, self.IL,
                    self.IL)
        dae.add_jac(Fy0, mul(self.u, self.iL), self.IL, self.v1)
        dae.add_jac(Fy0, -mul(self.u, self.iL), self.IL, self.v2)


class RCp(DCBase):
    """RC parallel line"""

    def __init__(self, system, name):
        super(RCp, self).__init__(system, name)
        self._name = 'RCp'
        self._params.extend(['R', 'C'])
        self._data.update({
            'R': 0.01,
            'C': 0.001,
        })
        self._params.extend(['R', 'C'])
        self._r.extend(['R'])
        self._g.extend(['C'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['vC'])
        self._fnamex.extend(['v_C'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._service.extend(['iR', 'iC'])
        self._init()

    def servcall(self, dae):
        self.iR = div(self.u, self.R)
        self.iC = div(self.u, self.C)

    def init0(self, dae):
        self.servcall(dae)
        dae.x[self.vC] = self.v12
        dae.y[self.Idc] = -mul(self.v12, self.iR)

    def gcall(self, dae):
        dae.g[self.Idc] = dae.x[self.vC] - self.v12
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        dae.f[self.vC] = -mul(dae.y[self.Idc] + mul(dae.x[self.vC], self.iR),
                              self.iC)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.vC)
        dae.add_jac(Gy0, -self.u, self.Idc, self.v1)
        dae.add_jac(Gy0, self.u, self.Idc, self.v2)
        dae.add_jac(Gy0, 1e-6, self.Idc, self.Idc)

        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)

        dae.add_jac(Fy0, -self.iC, self.vC, self.Idc)
        dae.add_jac(Fx0, -mul(self.iR, self.iC), self.vC, self.vC)


class RLCp(DCBase):
    """RLC parallel line"""

    def __init__(self, system, name):
        super(RLCp, self).__init__(system, name)
        self._name = 'RLCp'
        self._params.extend(['R', 'L', 'C'])
        self._data.update({
            'R': 0.01,
            'L': 0.001,
            'C': 0.001,
        })
        self._params.extend(['R', 'L', 'C'])
        self._r.extend(['R', 'L'])
        self._g.extend(['C'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['IL', 'vC'])
        self._fnamex.extend(['I_L', 'v_C'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._service.extend(['iR', 'iL', 'iC'])
        self._init()

    def servcall(self, dae):
        self.iR = div(self.u, self.R)
        self.iL = div(self.u, self.L)
        self.iC = div(self.u, self.C)

    def init0(self, dae):
        self.servcall(dae)
        dae.x[self.vC] = self.v12
        dae.y[self.Idc] = -mul(self.v12, self.iR)

    def gcall(self, dae):
        dae.g[self.Idc] = dae.x[self.vC] - self.v12
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        dae.f[self.IL] = mul(dae.x[self.vC], self.iL)
        dae.f[self.vC] = -mul(
            dae.y[self.Idc] + mul(dae.x[self.vC], self.iR) + dae.x[self.IL],
            self.iC)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.vC)
        dae.add_jac(Gy0, -self.u, self.Idc, self.v1)
        dae.add_jac(Gy0, self.u, self.Idc, self.v2)
        dae.add_jac(Gy0, 1e-6, self.Idc, self.Idc)

        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)

        dae.add_jac(Fx0, self.iL, self.IL, self.vC)

        dae.add_jac(Fy0, -self.iC, self.vC, self.Idc)
        dae.add_jac(Fx0, -mul(self.iR, self.iC), self.vC, self.vC)
        dae.add_jac(Fx0, -self.iC, self.vC, self.IL)


class RCs(DCBase):
    """RC series line"""

    def __init__(self, system, name):
        super(RCs, self).__init__(system, name)
        self._name = 'RCs'
        self._data.update({'R': 0.01, 'C': 0.001})
        self._params.extend(['R', 'C'])
        self._r.extend(['R'])
        self._g.extend(['C'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['vC'])
        self._fnamex.extend(['vC'])
        self._service.extend(['iC', 'iR'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._init()

    def servcall(self, dae):
        self.iC = div(self.u, self.C)
        self.iR = div(self.u, self.R)

    def init0(self, dae):
        self.servcall(dae)

    def gcall(self, dae):
        dae.g[self.Idc] = dae.x[self.vC] - self.v12 - mul(
            dae.y[self.Idc], self.R)
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        dae.f[self.vC] = -mul(dae.y[self.Idc], self.iC)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.vC)
        dae.add_jac(Gy0, -self.u, self.Idc, self.v1)
        dae.add_jac(Gy0, self.u, self.Idc, self.v2)
        dae.add_jac(Gy0, 1e-6 - self.R, self.Idc, self.Idc)

        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)

        dae.add_jac(Fy0, -self.iC, self.vC, self.Idc)


class RLCs(DCBase):
    """RLC series"""

    def __init__(self, system, name):
        super(RLCs, self).__init__(system, name)
        self._name = 'RLCs'
        self._params.extend(['R', 'L', 'C'])
        self._data.update({'R': 0.01, 'L': 0.001, 'C': 0.001})
        self._params.extend(['R', 'L', 'C'])
        self._r.extend(['R', 'L'])
        self._g.extend(['C'])
        self._algebs.extend(['Idc'])
        self._fnamey.extend(['I_{dc}'])
        self._states.extend(['IL', 'vC'])
        self._fnamex.extend(['I_L', 'v_C'])
        self.calls.update({
            'pflow': True,
            'init0': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._service.extend(['iR', 'iL', 'iC'])
        self._init()

    def servcall(self, dae):
        self.iR = div(self.u, self.R)
        self.iL = div(self.u, self.L)
        self.iC = div(self.u, self.C)

    def init0(self, dae):
        self.servcall(dae)
        dae.x[self.vC] = self.v12

    def gcall(self, dae):
        dae.g[self.Idc] = dae.x[self.IL] + dae.y[self.Idc]
        dae.g -= spmatrix(dae.y[self.Idc], self.v1, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(dae.y[self.Idc], self.v2, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        dae.f[self.IL] = mul(
            self.v12 - mul(self.R, dae.x[self.IL]) - dae.x[self.vC], self.iL)
        dae.f[self.vC] = mul(dae.x[self.IL], self.iC)

    def jac0(self, dae):
        dae.add_jac(Gx0, self.u, self.Idc, self.IL)
        dae.add_jac(Gy0, self.u, self.Idc, self.Idc)

        dae.add_jac(Gy0, -self.u, self.v1, self.Idc)
        dae.add_jac(Gy0, self.u, self.v2, self.Idc)

        dae.add_jac(Fx0, -mul(self.R, self.iL), self.IL, self.IL)
        dae.add_jac(Fy0, self.iL, self.IL, self.v1)
        dae.add_jac(Fy0, -self.iL, self.IL, self.v2)
        dae.add_jac(Fx0, -self.iL, self.IL, self.vC)

        dae.add_jac(Fx0, self.iC, self.vC, self.IL)


class Ground(DCBase):
    """DC Ground node"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self.param_remove('node1')
        self.param_remove('node2')
        self.param_remove('v')
        self._data.update({
            'node': None,
            'voltage': 0.0,
        })
        self._algebs.extend(['I'])
        self._unamey = ['I']
        self._fnamey = ['I']

        self._dc = {'node': 'v'}
        self._params.extend(['voltage'])
        self._mandatory.extend(['node'])
        self.calls.update({
            'init0': True,
            'jac0': True,
            'gcall': True,
            'pflow': True,
        })

        self._init()

    def init0(self, dae):
        dae.y[self.v] = self.voltage

    def gcall(self, dae):
        dae.g[self.v] -= dae.y[self.I]
        dae.g[self.I] = self.voltage - dae.y[self.v]

    def jac0(self, dae):
        dae.add_jac(Gy0, -self.u, self.v, self.I)
        dae.add_jac(Gy0, self.u - 1 - 1e-6, self.I, self.I)
        dae.add_jac(Gy0, -self.u, self.I, self.v)


class DCgen(DCBase):
    """A simple DC generator to impose constantactive power injection"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'DCgen'
        self._params.extend(['P', 'Sn'])
        self._data.update({'P': 0.0,
                           })
        self._descr.update({'P': 'Active power generation'})
        self._units.update({'P': 'pu'})
        self._powers.extend(['P'])
        self.calls.update({'pflow': True,
                           'gcall': True,
                           'stagen': True,
                           })
        self._init()

    def gcall(self, dae):
        dae.g -= spmatrix(div(mul(self.u, self.P), self.v12), self.v1, [0] * self.n, (dae.m, 1), 'd')
        dae.g -= spmatrix(-div(mul(self.u, self.P), self.v12), self.v2, [0] * self.n, (dae.m, 1), 'd')

    def disable_gen(self, idx):
        self.u[self.uid[idx]] = 0
        self.system.dae.factorize = True


class DCload(DCBase):
    """A simple DC load to impose constant active power consumption"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'DCload'
        self._params.extend(['P', 'Sn'])
        self._data.update({'P': 0.0,
                           })
        self._powers.extend(['P'])
        self._descr.update({'P': 'Active power load'})
        self._units.update({'P': 'pu'})
        self.calls.update({'pflow': True,
                           'gcall': True,
                           })
        self._init()

    def gcall(self, dae):
        dae.g += spmatrix(div(mul(self.u, self.P), self.v12), self.v1, [0] * self.n, (dae.m, 1), 'd')
        dae.g += spmatrix(-div(mul(self.u, self.P), self.v12), self.v2, [0] * self.n, (dae.m, 1), 'd')

    def disable(self, idx):
        self.u[self.uid[idx]] = 0
        self.system.dae.factorize = True



import sys
import logging
import numpy as np
import importlib
import platform

from cvxopt import mul, div, matrix, sparse, spdiag, spmatrix
from cvxopt.modeling import variable, op  # NOQA
from andes.consts import Gx, Fy0, Gy0
from andes.models.base import ModelBase
from andes.utils.math import zeros, index
from andes.utils.solver import Solver

logger = logging.getLogger(__name__)
cp = None


class BArea(ModelBase):
    """
    Balancing area class. This class defines power balancing area on top of the `Area` class for calculating
    center of inertia frequency, total inertia, expected power and area control error.
    """
    def __init__(self, system, name):
        super(BArea, self).__init__(system, name)
        self._group = 'Calculation'
        self._data.update({
            'area': None,
            'syn': None,
            'beta': 0,
        })
        self._descr.update({'area': 'Idx of Area',
                            'beta': 'Beta coefficient to multiply by the pu freq. deviation',
                            'syn': 'Indices of generators for computing COI'
                            })
        self._units.update({'syn': 'list'})
        self._mandatory.extend(['area', 'syn', 'beta'])
        self._algebs.extend(['Pexp', 'fcoi', 'ace'])
        self.calls.update({
            'gcall': True,
            'init1': True,
            'jac0': True,
        })
        self._service.extend(['P0', 'Mtot', 'M', 'usyn', 'wsyn'])
        self._fnamey.extend(['P_{exp}', 'f_{coi}', 'ace'])
        self._params.extend(['beta'])
        self._init()

    def init1(self, dae):

        for item in self._service:
            self.__dict__[item] = [[]] * self.n

        # Start with frequency
        for idx, item in enumerate(self.syn):
            self.M[idx] = self.read_data_ext('Synchronous', field='M', idx=item)
            self.Mtot[idx] = sum(self.M[idx])
            self.usyn[idx] = self.read_data_ext('Synchronous', field='u', idx=item)
            self.wsyn[idx] = self.read_data_ext('Synchronous', field='omega', idx=item)
            dae.y[self.fcoi[idx]] = sum(mul(self.M[idx], dae.x[self.wsyn[idx]])) / self.Mtot[idx]

        # Get BA Export Power
        self.copy_data_ext('Area', field='area_P0', dest='P0', idx=self.area)
        dae.y[self.Pexp] = self.P0
        dae.y[self.ace] = 0

    def gcall(self, dae):

        # the value below gets updated at each iteration in `seriesflow`
        P = self.read_data_ext('Area', field='area_P0', idx=self.area)
        dae.g[self.Pexp] = dae.y[self.Pexp] - P

        for idx, item in enumerate(self.syn):
            self.wsyn[idx] = self.read_data_ext('Synchronous', field='omega', idx=item)
            dae.g[self.fcoi[idx]] = dae.y[self.fcoi[idx]] - \
                sum(mul(self.M[idx], dae.x[self.wsyn[idx]])) / self.Mtot[idx]

        ACE = (P - self.P0) - mul(self.beta, (1 - dae.y[self.fcoi]))

        dae.g[self.ace] = dae.y[self.ace] + ACE

    def jac0(self, dae):
        dae.add_jac(Gy0, 1, self.Pexp, self.Pexp)
        dae.add_jac(Gy0, 1, self.fcoi, self.fcoi)
        dae.add_jac(Gy0, 1, self.ace, self.ace)
        dae.add_jac(Gy0, 1, self.ace, self.Pexp)
        dae.add_jac(Gy0, self.beta, self.ace, self.fcoi)


class AGCBase(ModelBase):
    """
    Base AGC class. The allocation of Pagc will be based on inverse droop (iR)
    """
    def __init__(self, system, name):
        super(AGCBase, self).__init__(system, name)
        self._group = 'AGCGroup'
        self._data.update({'BArea': None,
                           'Ki': 0.05,
                           })
        self._descr.update({'BArea': 'Idx of BArea',
                            'Ki': 'Integral gain of ACE',
                            })
        self._mandatory.extend(['BArea', 'Ki'])
        self._states.extend(['Pagc'])
        self.calls.update({'init1': True,
                           'gcall': True,
                           'fcall': True,
                           'jac0': True,
                           'gycall': True
                           })
        self._service.extend(['ace', 'iR', 'iRtot'])
        self._fnamex.extend(['P_{agc}^{total}'])
        self._params.extend(['Ki'])

    def init1(self, dae):
        self.copy_data_ext('BArea', field='ace', idx=self.BArea)

    def fcall(self, dae):
        dae.f[self.Pagc] = mul(self.Ki, dae.y[self.ace])

    def gcall(self, dae):
        pass

    def jac0(self, dae):
        dae.add_jac(Fy0, self.Ki, self.Pagc, self.ace)

    def gycall(self, dae):
        pass


class AGCSyn(AGCBase):
    """AGC for synchronous generators. This class changes the setpoints by modifying the generator pm."""
    def __init__(self, system, name):
        super(AGCSyn, self).__init__(system, name)
        self._data.update({'syn': None})
        self._descr.update({'syn': 'Indices of synchronous generators for AGC'})
        self._units.update({'syn': 'list'})
        self._mandatory.extend(['syn'])
        self._service.extend(['pm', 'usyn'])
        self._init()

    def init1(self, dae):
        super(AGCSyn, self).init1(dae)
        self.pm = [[]] * self.n
        self.iR = [[]] * self.n
        self.usyn = [[]] * self.n
        self.iRtot = [[]] * self.n

        for idx, item in enumerate(self.syn):
            self.pm[idx] = self.read_data_ext('Synchronous', field='pm', idx=item)
            self.usyn[idx] = self.read_data_ext('Synchronous', field='u', idx=item)
            self.iR[idx] = self.read_data_ext('Synchronous', field='M', idx=item)
            self.iRtot[idx] = sum(mul(self.usyn[idx], self.iR[idx]))

    def gcall(self, dae):
        super(AGCSyn, self).gcall(dae)

        # Kgen and each item in `self.pm`, `self.usyn`, and `self.Pagc` is a list
        #   Do not get rid of the `for` loop, since each of them is a matrix operation

        for idx, item in enumerate(self.syn):
            Kgen = div(self.iR[idx], self.iRtot[idx])
            dae.g[self.pm[idx]] -= mul(self.usyn[idx], Kgen, dae.x[self.Pagc[idx]])

    def gycall(self, dae):
        super(AGCSyn, self).gycall(dae)

        # Do not get rid of the for loop; for each `idx` it is a matrix operation

        for idx, item in enumerate(self.syn):
            Kgen = div(self.iR[idx], self.iRtot[idx])
            dae.add_jac(Gx, -mul(self.usyn[idx], Kgen), self.pm[idx], self.Pagc[idx])


class AGC(AGCSyn):
    """Alias for class <AGCSyn>"""
    pass


class AGCTG(AGCBase):
    """AGC class that modifies the turbine governor power reference. Links to TG1 only."""
    def __init__(self, system, name):
        super(AGCTG, self).__init__(system, name)
        self._data.update({'tg': None})
        self._mandatory.extend(['tg'])
        self._descr.update({'tg': 'Indices of turbine governors for AGC'})
        self._units.update({'tg': 'list'})
        self._service.extend(['pin', 'R', 'iR', 'iRtot'])

        self._init()

    def init1(self, dae):
        super(AGCTG, self).init1(dae)
        self.pin = [[]] * self.n
        self.R = [[]] * self.n
        self.iR = [[]] * self.n
        self.iRtot = [[]] * self.n

        for idx, item in enumerate(self.tg):
            self.pin[idx] = self.read_data_ext(model='Governor', field='pin', idx=item)
            self.R[idx] = self.read_data_ext(model='Governor', field='R', idx=item)
            self.iR[idx] = div(1, self.R[idx])
            self.iRtot[idx] = sum(self.iR[idx])

    def gcall(self, dae):
        super(AGCTG, self).gcall(dae)
        for idx, item in enumerate(self.tg):
            Ktg = div(self.iR[idx], self.iRtot[idx])
            dae.g[self.pin[idx]] += mul(Ktg, dae.x[self.Pagc[idx]])

    def gycall(self, dae):
        super(AGCTG, self).gycall(dae)
        for idx, item in enumerate(self.tg):
            Ktg = div(self.iR[idx], self.iRtot[idx])
            dae.add_jac(Gx, Ktg, self.pin[idx], self.Pagc[idx])


class AGCVSCBase(object):
    """
    Base class for AGC using VSC. Modifies the ref1 for PV or PQ-controlled VSCs. This class must be
    inherited with subclasses of AGCBase
    """
    def __init__(self, system, name):
        self.system = system
        self._data.update({'vsc': None,
                           'Rvsc': None,
                           })
        self._descr.update({'vsc': 'Indices of VSCs to control',
                            'Rvsc': 'Droop coefficients for the VSCs'})
        self._units.update({'tg': 'list',
                            'Rvsc': 'list'})
        self._mandatory.extend(['vsc', 'Rvsc'])
        self._service.extend(['uvsc', 'ref1'])
        self._init()

    def init1(self, dae):
        self.ref1 = [[]] * self.n
        self.uvsc = [[]] * self.n

        # manually convert self.Rvsc to a list of matrices
        self.Rvsc = [matrix(item) for item in self.Rvsc]
        self.iRvsc = [div(1, item) for item in self.Rvsc]

        # Only PV or PQ-controlled VSCs are acceptable
        for agc_idx, item in enumerate(self.vsc[:]):
            pv_or_pq = self.read_data_ext('VSCgroup', field="PV", idx=item) + \
                        self.read_data_ext('VSCgroup', field='PQ', idx=item)

            valid_vsc_list = list()
            valid_vsc_R = list()
            for i, (vsc_idx, valid) in enumerate(zip(item, pv_or_pq)):
                if valid:
                    valid_vsc_list.append(vsc_idx)
                    # TODO: fix the hard-coded `vsc_Idx` below
                    valid_vsc_R.append(self.Rvsc[agc_idx][i])
                else:
                    logger.warning('VSC <{}> is not a PV or PQ type, thus cannot be used for AGC.'.format(vsc_idx))
            self.vsc[agc_idx] = valid_vsc_list

        for agc_idx, item in enumerate(self.vsc):
            # skip elements that contain no valid VSC index
            if len(item) == 0:
                continue

            # retrieve status `uvsc`
            self.uvsc[agc_idx] = self.read_data_ext('VSCgroup', field='u', idx=item)
            self.ref1[agc_idx] = self.read_data_ext('VSCgroup', field='ref1', idx=item)
            # Add `Rvsc` to Mtot
            self.iRtot[agc_idx] += sum(mul(self.uvsc[agc_idx], self.iRvsc[agc_idx]))

    def gcall(self, dae):
        for agc_idx, item in enumerate(self.vsc):
            if len(item) == 0:
                continue

            Kvsc = div(self.iRvsc[agc_idx], self.iRtot[agc_idx])
            dae.g[self.ref1[agc_idx]] -= mul(self.uvsc[agc_idx], Kvsc, dae.x[self.Pagc[agc_idx]])

    def gycall(self, dae):

        for agc_idx, item in enumerate(self.vsc):
            if len(item) == 0:
                continue

            Kvsc = div(self.iRvsc[agc_idx], self.iRtot[agc_idx])
            dae.add_jac(Gx, -mul(self.uvsc[agc_idx], Kvsc), self.ref1[agc_idx], self.Pagc[agc_idx])


class AGCTGVSC(AGCTG, AGCVSCBase):
    """AGC class that modifies the turbine governor and VSC pref"""
    def __init__(self, system, name):
        AGCTG.__init__(self, system, name)
        AGCVSCBase.__init__(self, system, name)
        self._init()

    def init1(self, dae):
        AGCTG.init1(self, dae)
        AGCVSCBase.init1(self, dae)

    def jac0(self, dae):
        AGCTG.jac0(self, dae)

    def gcall(self, dae):
        AGCTG.gcall(self, dae)
        AGCVSCBase.gcall(self, dae)

    def gycall(self, dae):
        AGCTG.gycall(self, dae)
        AGCVSCBase.gycall(self, dae)

    def fcall(self, dae):
        AGCTG.fcall(self, dae)


class AGCMPC(ModelBase):
    """MPC based AGC using TG and VSC"""
    def __init__(self, system, name):
        super(AGCMPC, self).__init__(system, name)
        if platform.system() == 'Darwin':
            logger.error("** AGCMPC optimization does not work correctly on macOS!!!")

        self._group = "AGCGroup"
        self._name = "AGCMPC"
        self.param_remove('Vn')
        self.param_remove('Sn')

        self._data.update({'tg': None,
                           'avr': None,
                           'vsc': None,
                           'qw': 15000,
                           'qu': 10,
                           })
        self._params.extend(['qw', 'qu'])
        self._descr.update({'tg': 'idx for turbine governors',
                            'vsc': 'idx for VSC dynamic models',
                            'qw': 'the coeff for minimizing frequency deviation',
                            'qu': 'the coeff for minimizing input deviation'
                            })
        self._units.update({'tg': 'list', 'vsc': 'list'})
        self._mandatory.extend(['tg', 'avr'])
        self.calls.update({'init1': True,
                           'gcall': True,
                           'jac0': True,
                           'fxcall': True})

        self._service.extend(['xg10', 'pin0', 'delta0', 'omega0', 't', 'dpin0', 'x0', 'xlast'
                              'xidx', 'uidx', 'yxidx', 'sfx', 'sfu', 'sfy', 'sgx', 'sgu', 'sgy',
                              'A', 'B', 'Aa', 'Ba',
                              'obj', 'domega', 'du', 'dx', 'x', 'xpred'
                              'xa'])

        self._algebs.extend(['dpin'])
        self._fnamey.extend(r'\Delta P_{in}')
        self.solver = Solver(system.config.sparselib)
        self.H = 6
        self.uvar = None
        self.op = None
        self._linearized = False
        self._interval = 0  # AGC apply interval in seconds. 0 - continuous
        self._init()

    def init1(self, dae):
        if globals()['cp'] is None:
            try:
                globals()['cp'] = importlib.import_module('cvxpy')
            except ImportError:
                logger.error('CVXPY import error. Install optional package `cvxpy` to use AGCMPC')
                sys.exit(1)

        self.t = -1
        self.tlast = -1
        # state array x = [delta, omega, xg1]
        # input array u = [dpin]
        self.copy_data_ext('Governor', field='gen', dest='syn', idx=self.tg)
        self.copy_data_ext('Synchronous', field='delta', dest='delta', idx=self.syn)
        self.copy_data_ext('Synchronous', field='omega', dest='omega', idx=self.syn)
        self.copy_data_ext('Synchronous', field='e1d', dest='e1d', idx=self.syn)
        self.copy_data_ext('Synchronous', field='e1q', dest='e1q', idx=self.syn)
        self.copy_data_ext('Synchronous', field='e2d', dest='e2d', idx=self.syn)
        self.copy_data_ext('Synchronous', field='e2q', dest='e2q', idx=self.syn)

        self.copy_data_ext('Governor', field='xg1', dest='xg1', idx=self.tg)
        self.copy_data_ext('Governor', field='xg2', dest='xg2', idx=self.tg)
        self.copy_data_ext('Governor', field='xg3', dest='xg3', idx=self.tg)
        self.copy_data_ext('Governor', field='pin', dest='pin', idx=self.tg)

        self.copy_data_ext('AVR', field='vm', dest='vm', idx=self.avr)
        self.copy_data_ext('AVR', field='vr1', dest='vr1', idx=self.avr)
        self.copy_data_ext('AVR', field='vr2', dest='vr2', idx=self.avr)
        self.copy_data_ext('AVR', field='vfout', dest='vfout', idx=self.avr)

        dae.y[self.dpin] = 0
        self.dpin0 = zeros(self.n, 1)

        # build state/ input /other algebraic idx array
        self.xidx = matrix([self.delta, self.omega, self.e1d, self.e1q, self.e2d, self.e2q, self.xg1, self.xg2,
                            self.xg3, self.vm, self.vr1, self.vr2, self.vfout])
        self.x0 = dae.x[self.xidx]
        self.x = zeros(len(self.xidx), 1)
        self.dx = zeros(len(self.xidx), 1)
        self.xlast = dae.x[self.xidx]

        self.uidx = matrix([self.dpin])
        self.ulast = zeros(self.n, 1)
        self.dpin_calc = zeros(self.n, 1)

        self.widx = self.system.PQ.a
        self.w0 = self.system.PQ.p0
        self.wlast = matrix(self.w0)

        self.yidx = self.omega
        self.yidx_in_x = [index(self.xidx, y)[0] for y in self.yidx]
        yidx = np.delete(np.arange(dae.m), np.array(self.uidx))
        self.yxidx = matrix(yidx)

        # optimization problem
        self.uvar = cp.Variable((len(self.uidx), self.H+1), 'u')
        self.uzero = cp.Parameter((len(self.uidx), ), 'u0')
        self.xazero = cp.Parameter((2 * len(self.xidx), 1), 'xa')
        self.prob = None

        self.t_store = []
        self.xpred_store = []

    def gcall(self, dae):

        if self.t == -1:
            self.t = dae.t
            return

        if not self._linearized:
            # update the linearization points
            self._linearized = True
            self.t = dae.t
            self.tlast = dae.t
            self.sfx = dae.Fx[self.xidx, self.xidx]
            self.sfu = dae.Fy[self.xidx, self.uidx]
            self.sfy = dae.Fy[self.xidx, self.yxidx]

            self.sgx = dae.Gx[self.yxidx, self.xidx]
            self.sgu = dae.Gy[self.yxidx, self.uidx]
            self.sgw = spmatrix(1, self.widx, list(range(len(self.widx))), (len(self.yxidx), len(self.widx)))
            self.sgy = dae.Gy[self.yxidx, self.yxidx]

            # create state matrices
            self.gyigx = matrix(self.sgx)
            self.gyigu = matrix(self.sgu)
            self.gyigw = matrix(self.sgw)

            self.solver.linsolve(self.sgy, self.gyigx)
            self.solver.linsolve(self.sgy, self.gyigu)
            self.solver.linsolve(self.sgy, self.gyigw)

            self.A = (self.sfx - self.sfy * self.gyigx)
            self.B = (self.sfu - self.sfy * self.gyigu)
            self.C = -(self.sfy * self.gyigw)

            self.A = self.system.tds.h * self.A
            self.Aa = sparse([[self.A, self.A],
                              [spmatrix([], [], [], (self.A.size[0], self.A.size[1])),
                               spdiag([1] * len(self.xidx))]])
            self.Ba = sparse([self.B, self.B])
            self.Ca = sparse([self.C, self.C])

            # formulate optimization problem
            nx = len(self.xidx)
            nu = len(self.uidx)
            obj_x = 0

            xa_0 = self.xazero
            for i in range(self.H):
                # calculate Xa for each step in horizon H
                # du = cp.reshape(self.uvar[:, i+1], (nu, 1)) - self.uvar[:,i]
                du = cp.reshape(self.uvar[:, i+1] - self.uvar[:, i], (nu, 1))
                xa_i = matrix(self.Aa) * xa_0 + matrix(self.Ba) * du
                obj_x += cp.multiply(self.qw, cp.square(xa_i[nx:][self.yidx_in_x] - self.x0[self.yidx_in_x]))
                xa_0 = xa_i

            # construct the optimization problem
            self.obj_x = cp.sum(obj_x)

            self.obj_u = 0

            self.obj_u += cp.sum(
                cp.multiply(
                    np.array(self.qu).reshape((nu, )),
                    cp.sum(cp.square(self.uvar[:, 1:] - self.uvar[:, :-1]), axis=1)
                )
            )

            constraints = [self.uvar[:, 0] == self.uzero,
                           self.uvar[:, 1:] - self.uvar[:, :-1] <= 0.5,
                           self.uvar[:, 1:] - self.uvar[:, :-1] >= -0.5
                           ]

            self.prob = cp.Problem(cp.Minimize(self.obj_x + self.obj_u), constraints)

        if dae.t != self.t:
            self.t = dae.t
            nx = len(self.xidx)
            nu = len(self.uidx)

            # # update Delta x and x for current step
            self.x = dae.x[self.xidx]
            self.dx = self.x - self.xlast
            self.xa = matrix([self.dx, self.x])

            # assign values to self.uzero and self.xazero
            self.uzero.value = np.array(self.ulast).reshape((-1, ))
            self.xazero.value = np.array(self.xa).reshape((-1, 1))

            # use warm_start when possible
            if dae.t == 0:
                self.prob.solve()
            else:
                self.prob.solve(warm_start=1)

            self.dpin_calc = matrix(self.uvar.value[:, 1])

            # update every interval
            if (self.t - self.tlast) >= self._interval:
                self.tlast = self.t
                self.dpin0 = self.dpin_calc

            opt_val = self.prob.solution.opt_val
            logger.debug("t={:.4f}, obj={:.6f}, u[0]={:.6f}".format(dae.t, opt_val, self.uvar.value[0, 0]))

            self.t_store.append(self.t)
            xa_post = matrix(self.Aa) * self.xa + matrix(self.Ba) * (matrix(self.uvar.value[:, 0]) - self.ulast)
            self.xpred_store.append(xa_post[nx:][self.yidx_in_x][0])

            # # post-optimization evaluator
            # # u_val = matrix([[0, 0], [0, 0], [0, 0]])
            # u_val = matrix(self.uvar.value)
            # u_val = zeros(2, self.H)
            # obj_x = 0
            # xa_0 = self.xa
            # u_0 = self.ulast
            # for i in range(self.H):
            #     # calculate Xa for each step in horizon H
            #     du = np.reshape(u_val[:, i], (-1, 1)) - u_0
            #     xa_i = matrix(self.Aa) * xa_0 + matrix(self.Ba) * matrix(du) #+ matrix(self.Ca) * self.dw
            #     obj_x += mul(self.qw, (xa_i[nx:][self.yidx_in_x] - self.x0[self.yidx_in_x]) ** 2)
            #     xa_0 = xa_i
            #     u_0 = np.reshape(u_val[:, i], (-1, 1))
            # self.obj_x = sum(obj_x)
            # u2 = np.array(mul(u_val, u_val))
            # self.obj_u = sum(mul(self.qu, matrix(np.sum(u2, 1))))
            #
            # eval_obj = self.obj_x + self.obj_u
            # print("Post eval, t={:.4f} obj = {:.6f}, u = {:.6f}, {:.6f}".format(self.t, eval_obj, u_val[0, 0],
            #                                                                     u_val[1, 0]))
            # print("    obj_x = {}, obj_u = {}".format(self.obj_x, self.obj_u))

            # record data for the current step
            self.ulast = self.dpin_calc
            self.xlast = dae.x[self.xidx]

        dae.g[self.dpin] = dae.y[self.dpin] - self.dpin0
        dae.g[self.pin] += dae.y[self.dpin]  # positive `dpin` increases the `pin` reference

    def jac0(self, dae):
        dae.add_jac(Gy0, 1, self.dpin, self.dpin)
        dae.add_jac(Gy0, 1, self.pin, self.dpin)


class AGCSynVSC(AGCSyn, AGCVSCBase):
    """AGC class that modifies Synchronous pm and VSC pref"""
    def __init__(self, system, name):
        AGCSyn.__init__(self, system, name)
        AGCVSCBase.__init__(self, system, name)
        self._init()

    def init1(self, dae):
        AGCSyn.init1(self, dae)
        AGCVSCBase.init1(self, dae)

    def jac0(self, dae):
        AGCSyn.jac0(self, dae)

    def gcall(self, dae):
        AGCSyn.gcall(self, dae)
        AGCVSCBase.gcall(self, dae)

    def gycall(self, dae):
        AGCSyn.gycall(self, dae)
        AGCVSCBase.gycall(self, dae)

    def fcall(self, dae):
        AGCSyn.fcall(self, dae)


class eAGC(ModelBase):
    def __init__(self, system, name):
        super(eAGC, self).__init__(system, name)
        self._group = 'Control'
        self._data.update({
            'cl': None,
            'tl': 0,
            'Pl': None,
            'BA': None,
        })
        self._descr.update({
            'cl': 'Loss sharing coefficient (vector)',
            'tl': 'Time of generator loss',
            'Pl': 'Loss of power generation in pu (vector)',
            'BA': 'Balancing Area that support the Gen loss',
        })
        self._mandatory.extend(['cl', 'tl', 'Pl', 'BA'])
        self.calls.update({
            'gcall': True,
            'init1': True,
            'jac0': False,
            'fcall': False,
        })
        self._service.extend(['ace', 'en'])
        self._params.extend(['cl', 'tl', 'Pl'])
        self._init()

    def init1(self, dae):
        self.ace = [[]] * self.n
        for idx, item in enumerate(self.BA):
            self.ace[idx] = self.read_data_ext('BArea', field='ace', idx=item)

        self.en = zeros(self.n, 1)

    def switch(self):
        """Switch if time for eAgc has come"""
        t = self.system.dae.t
        for idx in range(0, self.n):
            if t >= self.tl[idx]:
                if self.en[idx] == 0:
                    self.en[idx] = 1
                    logger.info('Extended ACE <{}> activated at t = {}.'.format(self.idx[idx], t))

    def gcall(self, dae):
        self.switch()

        for idx in range(0, self.n):
            dae.g[self.ace[idx]] -= mul(self.en[idx], self.cl[:, idx],
                                        self.Pl[idx])



from . import ConfigBase
from ..utils.cached import cached


class CPF(ConfigBase):
    def __init__(self, **kwargs):
        self.method = 'perpendicular intersection'
        self.single_slack = False
        self.reactive_limits = False
        self.nump = 1000
        self.mu_init = 1.0
        self.hopf = False
        self.step = 0.1
        super(CPF, self).__init__(**kwargs)

    @cached
    def config_descr(self):
        descriptions = {
            'method': 'method for CPF routine analysis',
            'single_slack': 'use single slack bus mode',
            'reactive_limits': 'consider reactive power limits',
        }
        return descriptions
